/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2004 Gianni Tedesco.
 *
 * HTTP URI Normalization routines.
 *
 * This is pretty slow right now because we go through the URI many
 * many times over, in future we need to optimize it for a single
 * pass, but that will make the code more complex.
 *
 * TODO:
 *  o Fix endian bugs in unicode stuff
 *  o Normalize UTF-16 (%00%xx)
 *  o Research effect of UTF-16 encodings
 *  o Reduce stack usage in n_path()
 *  o Optimize for a single pass
 *
 * Functions:
 *  o chartype - ctype-alike for HTTP
 *  o hexchar - convert a single hex character to an integer
 *  o n_assure - Make sure the buffer is big enough for this URI.
 *  o http_normalize_init - Allocate initial buffer
 *  o http_normalize_fini - Free buffer
 *  o n_pcount - count file path components
 *  o n_path - normalize file paths for /./ and /../
 *  o n_ucs4 - Convert a UTF-8 sequence to a UCS4 character
 *  o n_utf8 - Normalize UTF-8 characters
 *  o n_hex -  Normalize hex-encoded characters
 *  o n_split - Split the URI from the query string
 *  o http_normalize - Main entry point, non-reentrant, uses static buffer.
*/

#include <firestorm.h>
#include <strtouint.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include "http_normalize.h"

static u_int8_t *abuf;
static size_t abuflen;

struct tokenbucket __private http_tb = {
	.cost = (RATE_SEC/1),
	.burst = (RATE_SEC/1) * 10,
};

static struct generator nads_gen = init_generator("http_normalize", &http_tb);
static struct alert nads_traversal = {
	.alert = "Directory Traversal",
	.sid = 1, .rev = 0,
	.priority = 5,
	.gen = &nads_gen,
	.order = ALERT_BUILTIN,
};
static struct alert nads_bad_utf8 = {
	.alert = "Malformed UTF-8",
	.sid = 2, .rev = 0,
	.priority = 5,
	.gen = &nads_gen,
	.order = ALERT_BUILTIN,
};
static struct alert nads_bad_hex = {
	.alert = "Malformed Hex Encoding",
	.sid = 3, .rev = 0,
	.priority = 5,
	.gen = &nads_gen,
	.order = ALERT_BUILTIN,
};
static struct alert nads_nul = {
	.alert = "NUL Character Insertion Attempt",
	.sid = 4, .rev = 0,
	.priority = 5,
	.gen = &nads_gen,
	.order = ALERT_BUILTIN,
};

#define T_HEX (1<<7)
#define T_ALL (T_HEX)
static u_int8_t tbl[0x100]={
	['a'] T_HEX | 10,
	['b'] T_HEX | 11,
	['c'] T_HEX | 12,
	['d'] T_HEX | 13,
	['e'] T_HEX | 14,
	['f'] T_HEX | 15,
	['A'] T_HEX | 10,
	['B'] T_HEX | 11,
	['C'] T_HEX | 12,
	['D'] T_HEX | 13,
	['E'] T_HEX | 14,
	['F'] T_HEX | 15,
	['0'] T_HEX | 0,
	['1'] T_HEX | 1,
	['2'] T_HEX | 2,
	['3'] T_HEX | 3,
	['4'] T_HEX | 4,
	['5'] T_HEX | 5,
	['6'] T_HEX | 6,
	['7'] T_HEX | 7,
	['8'] T_HEX | 8,
	['9'] T_HEX | 9,
};

static inline unsigned int chartype(u_int8_t c)
{
	return (unsigned int)tbl[c];
}

static inline u_int8_t hexchar(u_int8_t ch)
{
	return tbl[ch] & ~T_ALL;
}

static inline int n_assure(size_t len)
{
	u_int8_t *new;

	if ( likely(len <= abuflen) )
		return 1;
	
	new = realloc(abuf, len);
	if ( unlikely(new == NULL) )
		return 0;

	abuf = new;
	abuflen = len;
	return 1;
}

int __private http_normalize_init(void)
{
	/* Allocate early to get it mlocked in realtime mode */
	return n_assure(8192);
}

void __private http_normalize_fini(void)
{
	if ( abuf )
		free(abuf);
}

static inline int n_pcount(struct fvec *uri)
{
	size_t len = uri->len;
	u_int8_t *in, *out;
	u_int8_t *last = NULL;
	int cnt = 0;
	size_t i;

	for(in=out=uri->ptr, i = 0; i < len; i++, in++) {
		if ( nads_ws(NADS_WS_BACKSLASHES) && *in == '\\' )
			*in = '/';

		if ( *in == '/' ) {
			if ( last && last == in -1 ) {
				last = in;
				continue;
			}

			last = in;
			cnt++;
		}

		if ( nads_ws(NADS_WS_CASE_SENSITIVE) ) {
			*out++ = *in;
		}else{
			*out++ = tolower(*in);
		}
	}

	*out = '\0';
	uri->len = (size_t)(out - (u_int8_t *)uri->ptr);
	return cnt;
}

static int n_path(struct packet *p, struct fvec *uri)
{
	int x = n_pcount(uri);
	u_int8_t *c[x];
	size_t i, len;
	int in, out, o;
	u_int8_t *obuf;
	u_int8_t *str;

	len = uri->len;
	obuf = uri->ptr;

	/* Grab the directory components */
	for(i=o=0; i < len; i++) {
		if ( obuf[i] == '/' ) {
			obuf[i] = '\0';
			c[o++] = obuf + i + 1;
		}
	}

	/* Strip out /./ and /../ */
	for(in=out=0; in < x; in++) {
		if ( !strcmp(c[in], ".") )
			continue;

		if ( !strcmp(c[in], "..") ) {
			if ( --out == 0 ) {
				alert_tag(p, &nads_traversal, -1);
				return 0;
			}
			continue;
		}

		c[out++] = c[in];
	}

	/* Reconstruct the url */
	for(str = obuf + 1, x = 0; x < out; x++) {
		u_int8_t *tmp;

		for(tmp = c[x]; *tmp; *tmp++) {
			*str = *tmp;
			str++;
		}

		if ( x < out - 1 ) {
			*str = '/';
			str++;
		}
	}

	*obuf = '/';
	*str = '\0';

	uri->len = (size_t)(str - (u_int8_t *)uri->ptr);

	return 1;
}

static inline void n_ucs4(u_int32_t *n, u_int8_t *buf, int len)
{
	int s, x;
	u_int8_t fmask[6] = {
		0x01|0x02|0x04|0x08|0x10,
		0x01|0x02|0x04|0x08,
		0x01|0x02|0x04,
		0x01|0x02,
		0x01,
		0x0,
	};

	*n = 0;

	for(s=0, x = len; x-- > 0; s+=6) {
		u_int8_t mask;

		if ( x == 0 ) {
			mask = fmask[len - 2];
		}else{
			mask = 0x3f;
		}

		*n |= ((u_int32_t)(buf[x] & mask)) << s;
	}
}

static int n_utf8(struct packet *p, struct fvec *uri)
{
	size_t vlen = uri->len;
	int len=0, x=0;
	int state = 0;
	u_int8_t *out, *in;
	u_int8_t chr[6];
	u_int32_t ucs4;
	size_t i;

	if ( !nads_ws(NADS_WS_UTF8) )
		return 1;

	for(in=out=uri->ptr, i = 0; i < vlen; i++, in++) {

		if ( *in == 0xff || *in == 0xfe ) {
			alert_tag(p, &nads_bad_utf8, -1);
			return 0;
		}

		if( state == 0 ) {
			if ( *in & 0x80 ) {
				u_int8_t n = 0x40;

				for(len=1; n != 0x1; len++) {
					if ( (*in & n) == 0 )
						break;
					n >>= 1;
				}

				chr[0] = *in;
				state = 1;
				x = 1;
			}else{
				*out++ = *in;
			}
		}else{
			if ( (*in & 0xc0) != 0x80 ) {
#if 0
				/* FIXME: IIS Ignores continuation markers */
				alert_tag(p, &nads_bad_utf8, -1);
				return 0;
#endif
			}

			chr[x++] = *in;
			if ( x >= len ) {
				state = 0;
				n_ucs4(&ucs4, chr, len);
				/* FIXME: Shorten all overlong representations,
				 * not just the ASCII case...
				 */
				if ( ucs4 & 0xffffff00 ) {
					memcpy(out, chr, len);
					out += len;
				}else{
					*out++ = (u_int8_t)(ucs4 & 0xff);
				}
			}
		}
	}

	if ( state != 0 ) {
		alert_tag(p, &nads_bad_utf8, -1);
		return 0;
	}

	*out = '\0';
	uri->len = (size_t)(out - (u_int8_t *)uri->ptr);
	return 1;
}

static int n_hex(struct packet *p, struct fvec *uri)
{
	int state = 0;
	u_int8_t *out, *in;
	size_t len = uri->len;
	size_t i;
	size_t maxlen = 0, x = 0;
	u_int8_t hbuf[5];

	memset(hbuf, 0, 5);

	for(in=out=uri->ptr, i = 0; i < len; i++, in++) {
	switch(state) {
	case 0:
		if ( *in == '%' ) {
			if ( nads_ws(NADS_WS_MS_UTF16) &&
				( *(in+1) == 'u' || *(in+1) == 'U') ) {
				/* MS-UTF16 (%uXXXX) */
				maxlen = 2;
				in++;
				i++;
			}else{
				maxlen = 1;
			}
			state = 1;
			x = 0;
			break;
		}
		*out++ = *in;
		break;
	case 1:
		if ( chartype(*in) & T_HEX ) {
			hbuf[x++] = *in;
			if ( x >= (maxlen*2) ) {
				u_int8_t d1, d2;

				state = 0;

				d1 = hexchar(hbuf[0]) << 4;
				d1 |= hexchar(hbuf[1]);
				d2 = hexchar(hbuf[2])<<4;
				d2 |= hexchar(hbuf[3]);

				/* Treat %u00XX as %XX to make next check
				 * easier to perform.
				 */
				if ( maxlen == 2 && d1 == 0 ) {
					d1 = d2;
					maxlen = 1;
				}

				/* Don't insert NULs in to the stream */
				if ( d1 == 0 || (maxlen == 2  && d2 == 0) ) {
					alert_tag(p, &nads_nul, -1);
					return 0;
				}

				*out++ = d1;
				if ( maxlen == 2 )
					*out++ = d2;
			}
		}else if ( maxlen == 1 ) {
			/* IIS appears to ignore percent signs on their own
			 * followed by crap...
			 */
			if ( nads_ws(NADS_WS_PERCENT_HACK) && x == 0 ) {
				*out++ = '%';
				in--;
				i--;
				state = 0;
				break;
			}

			alert_tag(p, &nads_bad_hex, -1);
			return 0;
		}
		break;
	}
	}

	if ( state != 0 ) {
		alert_tag(p, &nads_bad_hex, -1);
		return 0;
	}

	*out = '\0';
	uri->len = (size_t)(out - (u_int8_t *)uri->ptr);
	return 1;
}

static void n_split(u_int8_t *buf, struct http_request *r)
{
	u_int8_t *ptr = r->uri.ptr;
	size_t len = r->uri.len;
	size_t i, out = 0;

	r->uri_path.ptr = abuf;
	r->uri_path.len = 0;

	/* Add a leading slash if it's missing */
	if ( ptr[0] != '/' )
		abuf[out++] = '/';

	for(i=0; i < len; i++) {
		if ( unlikely(ptr[i] == '?') ) {
			r->uri_query.ptr = abuf + (out+1);
			r->uri_query.len = len - (out+1);
			i++;
			break;
		}else{
			abuf[out++] = ptr[i];
		}
	}

	abuf[out++] = '\0';
	r->uri_path.len = out - 1;

	for(; i < len; i++)
		abuf[out++] = ptr[i];

	abuf[out] = '\0';

	if ( r->uri_query.len == 0 )
		r->uri_query.ptr = NULL;
}

void __private http_normalize(struct packet *p, struct http_request *r)
{
	struct tcp_session *s = p->layer[p->llen - 2].session;

	if ( !n_assure(r->uri.len + 2) )
		return;

	select_webserver(s->s_addr, r->port);

	/* Split out query string */
	n_split(abuf, r);

	/* Process the URI part */
	if ( !n_hex(p, &r->uri_path) )
		goto err;

	if ( nads_ws(NADS_WS_DOUBLE_HEX) && !n_hex(p, &r->uri_path) )
		goto err;

	if ( !n_utf8(p, &r->uri_path) )
		goto err;

	if ( !n_path(p, &r->uri_path) )
		goto err;

	/* If there is a query string, process that too */
	if ( r->uri_query.ptr == NULL )
		goto done;

	if ( !n_hex(p, &r->uri_query) )
		goto err_query;

	if ( !n_utf8(p, &r->uri_query) )
		goto err_query;

done:
#if 0
	mesg(M_DEBUG, "'%s' '%s'",
			(char *)r->uri_path.ptr,
			(char *)r->uri_query.ptr);
#endif
	return;

err:
	r->uri_path.ptr = NULL;
	r->uri_path.len = 0;
err_query:
	r->uri_query.ptr = NULL;
	r->uri_query.len = 0;
}
