/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include "tcpip.h"
#include <strtouint.h>

/* ipaddr (src,dst) matching */
struct ipaddr {
	struct ipaddr *next;
	u_int32_t addr;
	u_int32_t mask;
	int negate;
	struct alert *alert;
	struct detect *result;
};

struct ip_addr_detect {
	struct detect d;
	struct ipaddr *addrs;
};

static int parse_ipaddr(char *sstr, u_int32_t *addr, u_int32_t *mask)
{
	char *smask;
	struct in_addr tmp;
	char str[64];

	if ( snprintf(str, sizeof(str), "%s", sstr)<0 )
		return 0;

	if ( (smask=strchr(str, '/')) ) {
		*smask='\0';
		smask++;

		if ( strchr(smask, '.') ) {
			/* Dotted style mask notation (255.255.255.0) */
			if ( !inet_aton(smask, &tmp) )
				return 0;
			*mask = tmp.s_addr;
		}else{ /* CIDR notation (/24) */
			unsigned int bits;
			if ( strtouint(smask, &bits) )
				return 0;
			if ( bits > 32 || bits < 0 )
				return 0;
#if __BYTE_ORDER == __BIG_ENDIAN
			*mask=~(0xffffffff >> bits);
#elif __BYTE_ORDER == __LITTLE_ENDIAN
			*mask=~(0xffffffff << bits);
#else
#error "What in hells name?!"
#endif
		}
	}else{
		*mask=-1;
	}

	if ( !inet_aton(str, &tmp) )
		return 0;

	*addr=tmp.s_addr;

	*addr &= *mask;
	return 1;
}

static int ip_addr_validate(struct matcher *match, char *args,
				void **priv, struct modifier *m)
{
	struct ipaddr *p;

	if ( !args )
		return 0;

	if ( !(p=calloc(1, sizeof(*p))) ) {
		return 0;
	}

	if ( !parse_ipaddr(args, &p->addr, &p->mask) ) {
		free(p);
		return 0;
	}

	*priv=p;
	return 1;
}

static void ip_src_detect(struct detect *d, struct packet *pkt, int l)
{
	u_int32_t saddr = pkt->layer[l].h.ip->saddr;
	struct ip_addr_detect *x = (struct ip_addr_detect *)d;
	struct ipaddr *p;

	for(p=x->addrs; p; p=p->next) {
		if ( p->negate ^ ((saddr & p->mask) == p->addr) ) {
			if ( p->alert )
				alert_tag(pkt, p->alert, l);
			if ( p->result )
				p->result->match(p->result, pkt, l);
		}
	}

	if ( x->d.any )
		x->d.any->match(x->d.any, pkt, l);
}

static void ip_dst_detect(struct detect *d, struct packet *pkt, int l)
{
	u_int32_t daddr = pkt->layer[l].h.ip->daddr;
	struct ip_addr_detect *x = (struct ip_addr_detect *)d;
	struct ipaddr *p;

	for(p=x->addrs; p; p=p->next) {
		if ( p->negate ^ (((daddr & p->mask) == p->addr)) ) {
			if ( p->alert )
				alert_tag(pkt, p->alert, l);
			if ( p->result )
				p->result->match(p->result, pkt, l);
		}
	}

	if ( x->d.any )
		x->d.any->match(x->d.any, pkt, l);
}

static struct detect *ip_src_new(struct matcher *m)
{
	struct ip_addr_detect *d;

	d = calloc(1, sizeof(*d));
	if ( d == NULL )
		return NULL;

	d->d.match = ip_src_detect;

	return (struct detect *)d;
}

static struct detect *ip_dst_new(struct matcher *m)
{
	struct ip_addr_detect *d;

	d = calloc(1, sizeof(*d));
	if ( d == NULL )
		return NULL;

	d->d.match = ip_dst_detect;

	return (struct detect *)d;
}

static struct detect **ip_addr_insert(struct detect *d, struct sig_node *n)
{
	struct ipaddr *p = (struct ipaddr *)n->private;
	struct ip_addr_detect *x = (struct ip_addr_detect *)d;
	struct ipaddr *tmp;

	p->alert = n->alert;
	p->negate = n->negate;

	for(tmp=x->addrs; tmp; tmp=tmp->next) {
		if ( p->addr != tmp->addr )
			continue;
		if ( p->mask != tmp->mask )
			continue;
		if ( p->negate != tmp->negate )
			continue;

		tmp->alert = alert_best(tmp->alert, p->alert);
		n->match->free_priv(n->private);
		return &tmp->result;
	}

	p->next = x->addrs;
	x->addrs = p;

	return &p->result;
}

struct matcher ip_src_matcher = {
	.validate = ip_addr_validate,
	.free_priv = free,
	.new_node = ip_src_new,
	.insert_node = ip_addr_insert,
};

struct matcher ip_dst_matcher = {
	.validate = ip_addr_validate,
	.free_priv = free,
	.new_node = ip_dst_new,
	.insert_node = ip_addr_insert,
};
