/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <dotscara/fdctl.h>
#include <unistd.h>
#include <fcntl.h>

#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <exporter.h>
#include <plugin.h>

#define MAGIC_LBL 0xa1b2c3d4

/* If we don't have libpcap */
#ifndef lib_pcap_h
#define pcap_file_header	tcpd_file_header
#define pcap_pkthdr		tcpd_pkthdr

#define PCAP_VERSION_MAJOR	2
#define PCAP_VERSION_MINOR	4

struct tcpd_file_header{
	unsigned int	magic;
	unsigned short	version_major;
	unsigned short	version_minor;
	int		thiszone;
	unsigned int	sigfigs;
	unsigned int	snaplen;
	unsigned int	linktype;
};

struct tcpd_pkthdr {
	struct timeval	ts;
	unsigned int	caplen;
	unsigned int	len;
};
#endif

struct dump_priv {
	struct proto 	*proto;
	struct fobuf	buf;
	unsigned int	hdr;
	unsigned int	id;
};

/* Arguments */
static struct proto_child *find_proto(struct packet *pkt)
{
	struct proto_child *c;
	size_t i;
	int l;

	/* Find the first layer in the packet which
	 * contains a protocol usable by libpcap */
	for(l=0; l<pkt->llen; l++) {
		for(c=pcap_dlt->child,i=0; i < pcap_dlt->num_child; i++,c++) {
			if ( pkt->layer[l].proto==c->proto ) {
				return c;
			}
		}
	}

	return NULL;
}

static int set_proto(struct dump_priv *p, char *proto)
{
	struct proto_child *c;
	size_t i;

	/* Lookup a valid pcap protocol by name */
	for(c=pcap_dlt->child,i=0; i < pcap_dlt->num_child; i++,c++) {
		if ( !strcmp(proto, c->proto->name) ) {
			p->proto = c->proto;
			p->id = c->id;
			return 1;
		}
	}

	mesg(M_ERR,"dump: couldn't find libpcap protocol for '%s'", proto);
	return 0;
}

/* Actually write the file header to the file */
static int dump_fheader(struct dump_priv *priv)
{
	struct dump_priv *p=priv;
	struct pcap_file_header fh;
	int hs=sizeof(fh);

	fh.magic=MAGIC_LBL;
	fh.version_major=PCAP_VERSION_MAJOR;
	fh.version_minor=PCAP_VERSION_MINOR;
	fh.thiszone=0;
	fh.sigfigs=0;
	fh.snaplen=65535;
	fh.linktype=p->id;

	if ( !fobuf_write(&p->buf, &fh, hs) ) {
		mesg(M_CRIT,"dump: fobuf_write(): %s", sys_err2(NULL));
		return 2;
	}

	return 1;
}

/* FIXME: This could be structured much better */
static int dump_queue(struct sink *s, struct packet *pkt)
{
	struct dump_priv *p = s->priv;
	struct proto_child *c;
	struct pcap_pkthdr ph;
	unsigned int ofs;
	int layer;

	if ( !pkt )
		return 0;

	if ( p->proto && p->hdr == 1 )
		goto fast_path;

	/* The file is most likely corrupt by this point */
	if ( p->hdr == 2 ) {
		mesg(M_CRIT,"dump: warning: alert not recorded, "
				"previous I/O errors.");
		return 0;
	}

	/* Try find a protocol in order to write the
	 * pcap header */
	if ( !p->proto ) {
		if ( (c=find_proto(pkt)) ) {
			p->proto=c->proto;
			p->id=c->id;
			mesg(M_INFO,"dump: dumping in '%s' format (%u)",
				p->proto->name, p->id);
		}else{
			mesg(M_CRIT,"dump: alert not recorded, "
				"no suitable initial protocol.");
				return 1;
		}
	}

	/* If we haven't already written our header
	 * then do so */
	if ( !p->hdr ) {
		if ( (p->hdr=dump_fheader(p))==2 )
			return 0;
	}


fast_path:
	/* Find where to start dumping from */
	for(layer=0; layer<pkt->llen; layer++) {
		if ( pkt->layer[layer].proto == p->proto ) {
			goto dump_packet;
		}
	}

	mesg(M_CRIT,"dump: alert not recorded, "
		"no suitable protocol.");
	return 1;

dump_packet:
	/* Figure out how far in to the packet
	 * our first byte of data is */
	ofs=(unsigned int)(pkt->layer[layer].h.raw-pkt->base);

	/* Fill in the packet header */
	ph.caplen=pkt->caplen-ofs;
	ph.len=pkt->len-ofs;
	time_to_timeval(pkt->time, &ph.ts);

	/* Fill in the vectors */
	if ( !fobuf_write(&p->buf, &ph, sizeof(ph)) )
		return 0;
	if ( !fobuf_write(&p->buf, pkt->base+ofs, ph.caplen) )
		return 0;

	return 1;
}

static struct arg dump_args[]={
	{"file", ARGTYPE_STRING, .help = "Output filename"},
	{"proto", ARGTYPE_STRING, .help = "Base protocol"},
	{NULL,}
};

static int dump_init(struct sink *s, const char *args)
{
	struct dump_priv *p;
	char *fn;

	p = calloc(1, sizeof(*p));
	if ( p == NULL ) {
		mesg(M_ERR, "dump: calloc(): %s", sys_err());
		return 0;
	}

	switch ( args_parse(dump_args, args, p) ) {
	case -1:
		mesg(M_ERR,"dump: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	fn = arg_get_string(&dump_args[1].val.v_str);
	if ( fn && !set_proto(p, fn) )
		goto err;

	fn = arg_get_string(&dump_args[0].val.v_str);
	if ( dump_args[0].val.v_str.len ) {
		int fd;

		fd = open(fn, O_WRONLY|O_CREAT|O_TRUNC, 0600);
		if ( fd < 0 ) {
			mesg(M_ERR, "ascii: %s: open(): %s", fn, sys_err());
			goto err;
		}

		if ( !fobuf_attach(&p->buf, fd, 0, 0) ) {
			fd_close(fd);
			goto err;
		}
	}else{
		if ( !fobuf_attach(&p->buf, STDOUT_FILENO, 0, 0) )
			goto err;
	}

	s->priv = p;
	return 1;
err:
	free(p);
	return 0;
}

static int dump_free(struct sink *s)
{
	struct dump_priv *p = s->priv;
	int ret = 1;

	if ( !fobuf_close(&p->buf) ) {
		mesg(M_ERR, "fobuf_close(): %s", sys_err2(NULL));
		ret = 0;
	}

	free(p);
	return ret;
}

struct exporter firestorm_exporters[]={
	{
		.name = "dump",
		.desc =  "Libpcap (tcpdump) files",
		.queue = dump_queue,
		.init = dump_init,
		.free = dump_free,
		.args = dump_args,
	},
	{.name = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "exporter.dump",
	.desc		= "Log to a tcpdump file",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 0,
};
