/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
 *
 * TODO:
 *  o Implement non-buffered writeout path
 *  o Implement synchronous writeout path
 *  o double check do_dir_fsync is called where necessary
 *  o Support stormwall
 *  o Support different elog versions for crash recovery
*/

#undef DO_DIR_FSYNC

#if defined(__linux__)
#define _GNU_SOURCE 1
#endif

#include <firestorm.h>
#include <dotscara/fdctl.h>
#include <dotscara/fobuf.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#if HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif

#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <exporter.h>
#include <plugin.h>
#include <elog.h>

#if defined(__linux__) && defined(O_DIRECTORY)
#define DO_DIR_FSYNC 1
#endif

#define SPOOL_MAX_SUBPATH 41

static const char alertfn[]="/alert.elog";

struct spool_dir {
	struct list_head list;
	struct fobuf buf;
	off_t cur_size;
	timestamp_t initial_ts;
	ino_t inode;
#if DO_DIR_FSYNC 
	int dir_fd;
#endif
	char *alert;
	char *dir;
	char *dir_end;
	char pathdata[0];
};

struct spool_group {
	off_t max_size;
	timestamp_t timeout;
	struct notifier *hup;
	struct notifier *timer;
	struct list_head dirs;
};

#define ELOG_PKT_VEC 6
struct elog_pkt_vec {
	struct elog_pkthdr ph;
	struct fvec iov[ELOG_PKT_VEC];
	off_t tot_len;
	int io_len;
	char decode_buf[4096];
	char sbuf[512];
};
static struct elog_pkt_vec the_vec;

/* Pad the serialization buffer to 4 bytes and make sure
 * any padding is zeroed.
 */
static inline void pad_buffer(char *buf, size_t *len)
{
	size_t real_len = *len;

	real_len = (*len + 0x3) & ~0x3;

	memset(buf + *len, 0, real_len - *len);

	*len = real_len;
}

static int elog_write_pkt(struct elog_pkt_vec *vec,
				struct fobuf *fo,
				int buffer)
{
	int i;

#if HAVE_WRITEV
	if ( !buffer ) {
		int ret;
		int v = 0;
		off_t left = vec->tot_len;

		if ( !fobuf_flush(fo) )
			return 0;

again:
		ret = writev(fo->fd,
				(const struct iovec *)&vec->iov[v],
				vec->io_len);

		if ( likely((off_t)ret == left) )
			return 1;

		if ( ret < 0 ) {
			if ( errno == EINTR )
				goto again;
			return 0;
		}

		/* Skip over whats left */
		for(i=0; i < vec->io_len; i++) {
			if ( (size_t)ret >= vec->iov[v].len ) {
				v++;
				ret -= vec->iov[v].len;
				left -= vec->iov[v].len;
			}else{
				vec->iov[v].ptr += ret;
				vec->iov[v].len -= ret;
				left -= ret;
				goto again;
			}
		}
	}else
#endif
	for(i=0; i < vec->io_len; i++) {
		if ( !fobuf_write(fo, vec->iov[i].ptr, vec->iov[i].len) )
			return 0;
	}

#if !HAVE_WRITEV
	if ( !buffer && !fobuf_flush(fo) )
		return 0;
#endif

	return 1;
}

static int elog_build_pkt(struct elog_pkt_vec *vec, struct packet *pkt)
{
	static const char alert_padding[4];
	struct fvec *iov = vec->iov;
	struct elog_dhdr *dh;
	char *ptr = vec->decode_buf;
	char *ptr_end = vec->decode_buf + sizeof(vec->decode_buf);
	off_t tot_len = 0;
	char *gen_str, *alert_str;
	int i = 0;
	int j;

	/* Construct record header */
	vec->ph.h.type = ELOG_ALERT;
	if ( pkt->alert )
		vec->ph.h.prio = pkt->alert->priority;
	else
		vec->ph.h.prio = 0;
	vec->ph.h.reserved = 0;
	vec->ph.h.tv_sec = (pkt->time / TIMESTAMP_HZ);
	vec->ph.h.tv_usec = (pkt->time % TIMESTAMP_HZ) / TIMESTAMP_MHZ;

	/* Construct alert specific bit */
	if ( pkt->alert ) {
		vec->ph.sid = pkt->alert->sid;
		vec->ph.rev = pkt->alert->rev;
		vec->ph.gen_len = strlen(pkt->alert->gen->name) + 1;
		vec->ph.alert_len = strlen(pkt->alert->alert) + 1;
		gen_str = (char *)pkt->alert->gen->name;
		alert_str = (char *)pkt->alert->alert;
	}else{
		vec->ph.sid = 0;
		vec->ph.rev = 0;
		vec->ph.gen_len = 0;
		vec->ph.alert_len = 0;
		gen_str = NULL;
		alert_str = NULL;
	}
	vec->ph.pflags = (pkt->flags & FPMASK_PERSISTANT);
	vec->ph.decode_len = 0;
	vec->ph.pkt_len = pkt->len;
	vec->ph.pkt_caplen = pkt->caplen;

	/* Record header vector */
	iov[i].ptr = &vec->ph;
	iov[i].len = sizeof(vec->ph);
	tot_len += iov[i++].len;

	/* Construct the decode data */
	for(j=0; j<pkt->llen; j++) {
		int nlen, real, slen = 0;
		u_int32_t flags;

		if ( pkt->layer[j].proto ) {
			struct layer *l = &pkt->layer[j];

			/* Call plugins to serialise state data */
			if ( l->session && l->proto->serialize ) {
				slen = l->proto->serialize(pkt, j,
					vec->sbuf, sizeof(vec->sbuf));

				if ( slen < 0 ) {
					mesg(M_WARN,
						"alert: %s state data too big",
						l->proto->name);
					slen = 0;
				}else{
					pad_buffer(vec->sbuf, &slen);
				}
			}

			real = strlen(l->proto->name)+1;
			nlen = (real + 0x3) & ~0x3;
			flags = htonl(pkt->layer[j].flags);
		}else{
			real = nlen = slen = 0;
			flags = 0;
		}

		if ( ptr + nlen + sizeof(*dh) + slen > ptr_end ) {
			mesg(M_CRIT,"elog: decode buffer too small!");
			return 0;
		}

		/* Fill in the header */
		dh = (struct elog_dhdr *)ptr;
		dh->tot_len = (sizeof(*dh) + nlen + slen)>>2;
		dh->name_len = nlen >> 2;
		dh->pkt_ofs = htons(pkt->layer[j].h.raw - pkt->base);
		dh->flags = flags;
		ptr += sizeof(*dh);

		/* Then the protocol name  */
		if ( real ) {
			memcpy(ptr, pkt->layer[j].proto->name, real);
			ptr += real;

			/* make sure padding is zeroed */
			if ( nlen - real ) {
				memset(ptr, 0, nlen-real);
				ptr += nlen-real;
			}
		}

		/* then the session data */
		if ( slen ) {
			memcpy(ptr, vec->sbuf, slen);
			ptr += slen;
		}
	}

	/* decode data vector */
	iov[i].ptr = vec->decode_buf;
	iov[i].len = ptr - vec->decode_buf;

	/* fill in/update record header fields */
	vec->ph.decode_len = (iov[i].len >> 2);
	tot_len += iov[i++].len;

	/* raw packet data vector */
	iov[i].ptr = pkt->base;
	iov[i].len = pkt->caplen;
	tot_len += iov[i++].len;

	/* generator string */
	iov[i].ptr = gen_str;
	iov[i].len = vec->ph.gen_len;
	tot_len += iov[i++].len;

	/* Alert string */
	iov[i].ptr = alert_str;
	iov[i].len = vec->ph.alert_len;
	tot_len += iov[i++].len;

	/* Pad to multiples of 4 bytes (ensures
	 * aligned access to next header if file
	 * is mmap()ed in) */
	if ( tot_len & 3 ) {
		iov[i].ptr = (char *)alert_padding;
		iov[i].len = (4-(tot_len&3));
		tot_len += iov[i++].len;
	}

	/* Convert to network byte order  */
	vec->ph.h.reclen = htonl(tot_len);
	vec->ph.h.type = htons(vec->ph.h.type);
	vec->ph.h.tv_sec = htonl(vec->ph.h.tv_sec);
	vec->ph.h.tv_usec = htonl(vec->ph.h.tv_usec);
	vec->ph.sid = htonl(vec->ph.sid);
	vec->ph.rev = htonl(vec->ph.rev);
	vec->ph.pflags = htonl(vec->ph.pflags);
	vec->ph.decode_len = htons(vec->ph.decode_len);
	vec->ph.pkt_len = htonl(vec->ph.pkt_len);
	vec->ph.pkt_caplen = htonl(vec->ph.pkt_caplen);

	/* phew. this uses 5 or 6 of our I/O vectors, depending on
	 * whether or not we need to pad the end of the record. */
	vec->tot_len = tot_len;
	vec->io_len = i;
	return 1;
}

static struct arg elog_args[]={
	{"file", ARGTYPE_STRING, .help = "Output file"},
	{NULL,}
};

static int elog_queue(struct sink *s, struct packet *p)
{
	struct fobuf *o = s->priv;

	if ( !elog_build_pkt(&the_vec, p) ) {
		return 0;
	}

	if ( !elog_write_pkt(&the_vec, o, 1) ) {
		return 0;
	}

	return 1;
}

static int elog_init(struct sink *s, const char *args)
{
	struct efile_hdr fhdr;
	struct fobuf *f;
	char *fn;

	f = calloc(1, sizeof(*f));
	if ( f == NULL ) {
		mesg(M_ERR, "calloc(): %s", sys_err());
		return 0;
	}

	switch( args_parse(elog_args, args, NULL) ) {
	case -1:
		mesg(M_ERR, "elog: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	fn = arg_get_string(&elog_args[0].val.v_str);
	if ( elog_args[0].val.v_str.len ) {
		int fd;

		fd = open(fn, O_WRONLY|O_CREAT|O_TRUNC, 0600);
		if ( fd < 0 ) {
			mesg(M_ERR, "ascii: %s: open(): %s", fn, sys_err());
			goto err;
		}

		if ( !fobuf_attach(f, fd, 0, 0) ) {
			fd_close(fd);
			goto err;
		}
	}else{
		if ( !fobuf_attach(f, STDOUT_FILENO, 0, 0) )
			goto err;
	}

	fhdr.magic = htonl(EF_MAGIC);
	fhdr.flags = 0;
	fhdr.vers_minor = EF_VERS_MIN;
	fhdr.vers_major = EF_VERS_MAJ;
	if ( !fobuf_write(f, &fhdr, sizeof(fhdr)) ) {
		mesg(M_ERR, "fobuf_write() header: %s", sys_err2(NULL));
		goto err;
	}

	s->priv = f;
	return 1;

err:
	free(f);
	return 0;
}

static int elog_free(struct sink *s)
{
	struct fobuf *f = s->priv;
	int ret = 1;

	if ( !fobuf_close(f) ) {
		mesg(M_ERR, "fobuf_close(): %s", sys_err2(NULL));
		ret = 0;
	}

	free(f);

	return ret;
}

static int cb_dir(struct arg *a, struct fvec *v, void *user)
{
	struct list_head *h = user;
	struct spool_dir *d;
	size_t tot_len;

	if ( v == NULL )
		return 1;

	tot_len = sizeof(*d);
	tot_len += v->len + sizeof(alertfn);
	tot_len += v->len + SPOOL_MAX_SUBPATH;

	d = malloc(tot_len);
	if ( d == NULL )
		return 0;

	/* Do alert filename */
	d->alert = d->pathdata;
	memcpy(d->alert, v->ptr, v->len);
	memcpy(d->alert + v->len, alertfn, sizeof(alertfn));

	/* Do logdir/spoolname filename */
	d->dir = d->pathdata + v->len + sizeof(alertfn);
	memcpy(d->dir, v->ptr, v->len);
	d->dir_end = d->dir + v->len;
	*d->dir_end = '\0';

	/* for spool_isempty() check */
	d->initial_ts = 0;
	memset(&d->buf, 0, sizeof(d->buf));
	list_add_tail(&d->list, h);
	return 1;
}

static const char *spool_fn_alert(struct spool_dir *d)
{
	return d->alert;
}

static const char *spool_fn_spool(struct spool_dir *d)
{
	snprintf(d->dir_end,
		SPOOL_MAX_SUBPATH,
		"/@%.16llx.%.16llx.elog",
		d->initial_ts,
		(unsigned long long)d->inode);
	return d->dir;
}

static const char *spool_fn_dir(struct spool_dir *d)
{
	*d->dir_end = '\0';
	return d->dir;
}

static int do_dir_fsync(struct spool_dir *d)
{
	int ret = 1;

#if DO_DIR_FSYNC
	if ( fsync(d->dir_fd) && errno != EINVAL && errno != EROFS ) {
		mesg(M_ERR, "%s: fsync: %s", spool_fn_dir(d), sys_err());
		ret = 0;
	}
#endif

	/* TODO: wake up stormwall */

	return ret;
}

static int spool_creat(struct spool_dir *d);

static int spool_isempty(struct spool_dir *d)
{
	return !d->initial_ts;
}

static int spool_rotate(struct spool_dir *d, int new)
{
	int ret = 1;

	if ( spool_isempty(d) && new )
		return 1;

	if ( !fobuf_newfd(&d->buf, -1) )
		ret = 0;

	if ( !spool_isempty(d) ) {
		if ( rename(spool_fn_alert(d), spool_fn_spool(d)) )
			ret = 0;

		mesg(M_INFO, "elog-spool: logrotate: %s", spool_fn_spool(d));

		if ( !do_dir_fsync(d) )
			ret = 0;
	}else{
		const char *fn = spool_fn_alert(d);

		if ( unlink(fn) ) {
			mesg(M_ERR, "elog-spool: %s: unlink(): %s",
				fn, sys_err());
			ret = 0;
		}

		if ( !do_dir_fsync(d) )
			ret = 0;
	}
	
	if ( new && !spool_creat(d) )
		ret = 0;

	return ret;
}

static int spool_cond_rotate(struct spool_group *g, struct spool_dir *d,
				off_t len, timestamp_t ts)
{
	off_t tot_len = d->cur_size + len;
	timestamp_t rtime = d->initial_ts + g->timeout;

	/* Always allow 1 packet to be added */
	if ( spool_isempty(d) )
		return 1;

	if ( g->max_size && (tot_len > g->max_size) )
		return spool_rotate(d, 1);

	if ( g->timeout && !time_before(ts, rtime) )
		return spool_rotate(d, 1);

	return 1;
}

static int spool_dump_header(struct spool_dir *d)
{
	struct efile_hdr fhdr;
	size_t hsize = sizeof(fhdr);

	fhdr.magic = htonl(EF_MAGIC);
	fhdr.flags = 0;
	fhdr.vers_minor = EF_VERS_MIN;
	fhdr.vers_major = EF_VERS_MAJ;

	if ( !fobuf_write(&d->buf, &fhdr, hsize) ) {
		return 0;
	}

	d->cur_size += hsize;

	return 1;
}

static int spool_creat(struct spool_dir *d)
{
	struct stat st;
	const char *fn;
	int fd;

	fn = spool_fn_alert(d);
	fd = open(fn, O_WRONLY|O_CREAT|O_TRUNC, 00640);
	if ( fd < 0 ) {
		mesg(M_ERR, "elog-spool: %s: open(): %s", fn, sys_err());
		return 0;
	}

	if ( fstat(fd, &st) ) {
		mesg(M_ERR, "elog-spool: %s: fstat(): %s", fn, sys_err());
		fd_close(fd);
		return 0;
	}

	if ( !fobuf_newfd(&d->buf, fd) ) {
		mesg(M_ERR, "elog-spool: %s: fobuf_newfd(): %s", fn, sys_err());
		fd_close(fd);
		return 0;
	}

	d->inode = st.st_ino;
	d->cur_size = 0;
	d->initial_ts = 0;

	if ( !spool_dump_header(d) ) {
		mesg(M_ERR, "elog-spool: %s: spool_dump_header(): %s",
			fn, sys_err2(NULL));
		fobuf_close(&d->buf);
		return 0;
	}

	return 1;
}

static int spool_crash_recover(struct spool_dir *d);
static int spool_open(struct spool_dir *d, size_t bufsz)
{
	const char *fn = spool_fn_dir(d);

	if ( mkdir(fn, 0700) && errno != EEXIST ) {
		mesg(M_ERR, "elog-spool: %s: mkdir(): %s", fn, sys_err());
		return 0;
	}

#if DO_DIR_FSYNC
	d->dir_fd = open(fn, O_RDONLY|O_DIRECTORY);
	if ( d->dir_fd < 0 ) {
		mesg(M_ERR, "elog-spool: %s: open(): %s", fn, sys_err());
		return 0;
	}
#endif

	do_dir_fsync(d);

	if ( !spool_crash_recover(d) )
		goto err_close;

	if ( !fobuf_attach(&d->buf, -1, 0, bufsz) ) {
		mesg(M_ERR, "elog-spool: fobuf_attach(): %s",
			sys_err2(NULL));
		goto err_close;
	}

	if ( !spool_creat(d) )
		goto err_close;

	return 1;
err_close:
#if DO_DIR_FSYNC
	fd_close(d->dir_fd);
#endif
	return 0;
}

static int spool_close(struct spool_dir *d)
{
	do_dir_fsync(d);

#if DO_DIR_FSYNC
	fd_close(d->dir_fd);
#endif

	return 1;
}

static int spool_crash_recover(struct spool_dir *d)
{
	struct {
		struct efile_hdr fh;
		struct elog_common_hdr ch;
	}buf;
	int fd;
	size_t sz = sizeof(buf);
	int eof = 0;

	fd = open(spool_fn_alert(d), O_RDONLY);
	if ( fd < 0 ) {
		if ( errno == ENOENT )
			return 1;
		/* It exists, but we can't open it, that means
		 * we can't rotate it. Best die...
		 */
		goto err;
	}

	if ( !fd_read(fd, &buf, &sz, &eof) )
		goto err_close;

	if ( eof || sz != sizeof(buf) ||
		ntohl(buf.fh.magic) != EF_MAGIC ||
		ntohl(buf.fh.vers_major) != EF_VERS_MAJ ||
		ntohl(buf.fh.vers_minor) != EF_VERS_MIN ) {
		if ( unlink(spool_fn_alert(d)) )
			goto err_close;
		mesg(M_WARN, "elog-spool: obliterated %s",
			spool_fn_alert(d));
	}else{
		struct timeval tv;

		tv.tv_sec = ntohl(buf.ch.tv_sec);
		tv.tv_usec = ntohl(buf.ch.tv_usec);

		d->initial_ts = time_from_timeval(&tv);

		if ( rename(spool_fn_alert(d), spool_fn_spool(d)) )
			goto err_close;

		mesg(M_WARN, "elog-spool: crash-recover: %s",
				spool_fn_spool(d));

		if ( !do_dir_fsync(d) )
			goto err_close;
	}

	fd_close(fd);
	return 1;
err_close:
	fd_close(fd);
err:
	mesg(M_ERR, "%s: unable to perform crash recovery: %s",
			spool_fn_alert(d), sys_err());
	return 0;
}

static void spool_hup(struct packet *p, int layer, void *priv)
{
	struct spool_group *g = priv;
	struct spool_dir *d;

	list_for_each_entry(d, &g->dirs, list)
		spool_rotate(d, 1);
}

static void spool_timer(struct packet *p, int layer, void *priv)
{
	struct spool_group *g = priv;
	struct spool_dir *d;
	timestamp_t now;

	now = time_gettime();

	mesg(M_DEBUG, "Got timer signal");

	list_for_each_entry(d, &g->dirs, list)
		spool_cond_rotate(g, d, 0, now);
}

static timestamp_t ts;
static void itimer_mod(timestamp_t new_ts)
{
	struct itimerval it;
	unsigned int secs;

	ts = time_gcd(ts, new_ts);
	secs = ts / TIMESTAMP_HZ;

	time_to_timeval(ts, &it.it_interval);
	time_to_timeval(ts, &it.it_value);

	/* XXX: UNIX Specific */
	if ( setitimer(ITIMER_REAL, &it, NULL) ) {
		mesg(M_ERR, "setitimer(): %u seconds: %s", secs, sys_err());
		mesg(M_WARN, "Logs may not be rotated after max time");
	}else{
		if ( ts != 0 )
			mesg(M_INFO, "Using a %u second timer", secs);
	}
}

static struct arg spool_args[]={
	{"minutes", ARGTYPE_UINT, .help = "Rotation timeout"},
	{"size", ARGTYPE_OFF, .def = "1M", .help = "Maximum log size"},
	{"buf", ARGTYPE_BYTES, .def = "32K", .help = "Buffer size"},
	{"dir", ARGTYPE_CUSTOM, {.v_custom = cb_dir},
		.help = "Output directory"},
	{NULL,}
};

static int spool_init(struct sink *s, const char *args)
{
	struct spool_group *g = NULL;
	struct spool_dir *d, *t;
	struct list_head dirs = LIST_HEAD_INIT(dirs);
	unsigned int bufsz;

	switch ( args_parse(spool_args, args, &dirs) ) {
	case -1:
		mesg(M_ERR, "elog: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	g = malloc(sizeof(*g));
	if ( g == NULL )
		goto err;

	INIT_LIST_HEAD(&g->dirs);
	g->timeout = spool_args[0].val.v_uint * 60 * TIMESTAMP_HZ;
	g->max_size = spool_args[1].val.v_off;
	bufsz = spool_args[2].val.v_bytes;

	g->hup = notify(NOTIFY_HUP, spool_hup, NULL, g);
	if ( pipeline_is_live(s->owner) ) {
		g->timer = notify(NOTIFY_ITIMER, spool_timer, NULL, g);
		itimer_mod(g->timeout);
	}else{
		g->timer = NULL;
	}

	list_for_each_entry_safe(d, t, &dirs, list) {
		if ( !spool_open(d, bufsz) )
			goto err_close;
		list_move(&d->list, &g->dirs);
	}

	s->priv = g;
	return 1;

err_close:
	list_for_each_entry_safe(d, t, &g->dirs, list) {
		spool_close(d);
		list_del(&d->list);
		free(d);
	}
err:
	list_for_each_entry_safe(d, t, &dirs, list) {
		list_del(&d->list);
		free(d);
	}

	if ( g )
		free(g);

	return 0;
}

static int spool_free(struct sink *s)
{
	struct spool_group *g = s->priv;
	struct spool_dir *d, *t;
	int ret = 1;

	list_for_each_entry_safe(d, t, &g->dirs, list) {
		if ( !spool_rotate(d, 0) )
			ret = 0;
		spool_close(d);
		list_del(&d->list);
		free(d);
	}

	if ( g->hup )
		notifier_cancel(g->hup);
	if ( g->timer )
		notifier_cancel(g->timer);

	free(g);

	return ret;
}

static int spool_queue(struct sink *s, struct packet *p)
{
	struct spool_group *g = s->priv;
	struct spool_dir *d;
	
	if ( list_empty(&g->dirs) )
		return 1;

	d = list_entry(g->dirs.next, struct spool_dir, list);
	list_move_tail(&d->list, &g->dirs);

	if ( !elog_build_pkt(&the_vec, p) ) {
		return 0;
	}

	if ( !spool_cond_rotate(g, d, the_vec.tot_len, p->time) )
		return 0;

	if ( !elog_write_pkt(&the_vec, &d->buf, 1) ) {
		return 0;
	}

	if ( spool_isempty(d) )
		d->initial_ts = p->time;

	d->cur_size += the_vec.tot_len;

	return 1;
}

struct exporter firestorm_exporters[]={
	{
		.name = "elog",
		.desc = "Firestorm extended log (elog) files.",
		.init = elog_init,
		.free = elog_free,
		.queue = elog_queue,
		.args = elog_args,
	},
	{
		.name = "elog-spool",
		.desc = "Reliable ELOG spooling with crash recovery",
		.init = spool_init,
		.free = spool_free,
		.queue = spool_queue,
		.args = spool_args,
	},
	{.name = NULL,}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "exporter.elog",
	.desc		= "Firestorm exetended logs (elog)",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 0,
	.ver_minor	= 1,
};
