/*
 * This file is part of firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco
 * Copyright (C) 1998-2002 Martin Roesch <roesch@sourcefire.com>
 *
 * This code was originally taken from snort, but nearly all of that
 * has gone apart from the boyer moore inner loop. I have removed the
 * delta2 heuristic to keep O(1) memory complexity without really
 * sacrificing performance. Original snort copyright notice is
 * reproduced above anyway.
 *
 * Code could be improved by reversing compare order and combining with
 * Knuth-Morris-Pratt heuristic.
 *
 * Knuth also postulated that this code could be improved by using a
 * larger alphabet with characters >255 representing sequences of
 * characters.
 *
 * TODO
 *  o Faster setwise string matching code
 */

#include <firestorm.h>

#include <errno.h>

#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <idx.h>
#include <field.h>
#include <matcher.h>
#include <plugin.h>
#include <detect.h>
#include <strtouint.h>
#include "mwm.h"

static struct proto *http_proto=NULL;

/* Traditional string matching data */
struct content_priv {
	char *string;
	int len;
	int depth;
	int offset;
	int flag_uri;
	int (*type)(struct content_priv *, char *, int);
	int *skip;
	int negate;
	struct alert *alert;
	struct detect *result;
	struct content_priv *next;
};

struct content_detect {
	struct detect d;
	void *pat;
	struct content_priv *raw;
	int id;
};

/* Build delta1 (skip table) - alphabet sized lookup table for boyer-moore */
static void bm_skip(char *x, int plen, int *skip)
{
	int *sptr = &skip[256];

	while( sptr-- != skip )
		*sptr=plen+1;

	while(plen != 0)
		skip[(unsigned char)*x++] = plen--;
}

static int str_match_regex(struct content_priv *sp, char *hs, int hlen)
{
	int skip_stride,shift_stride, p_idx;
	int literal=0;
	int regexcomp=0;
	int b_idx;

	/* Do the search */
	b_idx=sp->len;
	while(b_idx <= hlen ) {
		p_idx=sp->len;

		while(hs[--b_idx]==sp->string[--p_idx]
			|| (sp->string[p_idx]=='?' && !literal)
			|| (sp->string[p_idx]=='*' && !literal)
			|| (sp->string[p_idx]=='\\' && !literal) ) {

			if ( literal )
				literal=0;

			if ( !literal && sp->string[p_idx]=='\\' )
				literal=1;

			if ( sp->string[p_idx]=='*' ) {
				while(p_idx!=0 && sp->string[--p_idx]=='*');

				while(hs[--b_idx]!=sp->string[p_idx]) {
					regexcomp++;
					if ( b_idx==0 ) return 0;
				}
			}

			if (p_idx==0) return 1;
			if (b_idx==0) break;
		}

		skip_stride=sp->skip[(unsigned char)hs[b_idx]];
		shift_stride=(sp->len-p_idx)+1;

		/* micro-optimised max() function */
		b_idx += ( (skip_stride-shift_stride)>0 )
			? skip_stride : shift_stride;
		b_idx += regexcomp;
		regexcomp=0;
	}

	return 0;
}

static int str_match(struct content_priv *sp, char *hs, int hlen)
{
	int skip_stride, shift_stride, p_idx;
	int b_idx;

	/* Do the search */
	b_idx=sp->len;
	while(b_idx <= hlen) {
		p_idx=sp->len;

		while((unsigned char)hs[--b_idx]==
			(unsigned char)sp->string[--p_idx]) {
			if (b_idx<0) return 0;
			if (p_idx==0) return 1;
		}

		skip_stride=sp->skip[(unsigned char)hs[b_idx]];
		shift_stride=(sp->len-p_idx)+1;

		/* micro-optimised max() function */
		b_idx += ( (skip_stride-shift_stride)>0 )
			? skip_stride : shift_stride;
	}

	return 0;
}

static int str_match_nocase(struct content_priv *sp, char *hs, int hlen)
{
	int skip_stride, shift_stride, p_idx;
	int b_idx;

	/* Do the search */
	b_idx=sp->len;
	while(b_idx <= hlen ) {
		p_idx=sp->len;

		while(toupper((unsigned char)hs[--b_idx])==
			(unsigned char)sp->string[--p_idx]) {
			if (b_idx<0) return 0;
			if (p_idx==0) return 1;
		}

		skip_stride=sp->skip[toupper((unsigned char)hs[b_idx])];
		shift_stride=(sp->len-p_idx)+1;

		/* micro-optimised max() function */
		b_idx += ( (skip_stride-shift_stride)>0 )
			? skip_stride : shift_stride;
	}

	return 0;
}

static int hextouint(const char *s, unsigned int *v)
{
	unsigned long int val;
	char *end;

	val = strtoul(s, &end, 16);
	if ( end == s || *end != '\0' )
		return 0;
	if ( val == ULONG_MAX && errno == ERANGE )
		return 0;
	if ( val > UINT_MAX )
		return 0;
	if ( val < 0 )
		return 0;

	*v = val;

	return 1;
}

static int str_validate(char *args, void **priv, struct modifier *m, int mode)
{
	char *t;
	int state=0;
	char bin[3]={0,0,0};
	int i=0, j=0;
	struct content_priv *p;
	struct modifier *mod;
	int (*mfunc)(struct content_priv *, char *, int)=str_match;
	int escaped=0;

	if ( !args )
		return 0;

	/* zero length string is pointless */
	if ( !(j=strlen(args)) )
		return 0;

	/* We can survive without this */
	if ( !http_proto )
		http_proto=decode_proto("http");

	if ( !(p=calloc(1, sizeof(*p))) )
		return 0;

	for(mod=m; mod; mod=mod->next) {
		if ( !strcmp("depth", mod->name) ) {
			if ( strtouint(mod->args, &p->depth) ) {
				free(p);
				return 0;
			}
		}else if ( !strcmp("offset", mod->name) ) {
			if ( strtouint(mod->args, &p->offset) ) {
				free(p);
				return 0;
			}

			/* XXX: Hrm, is this a bug for bug compatilbility
			 * of a bug somewhere else, very strange... */
			if ( p->offset )
				p->offset--;
		}else if ( !strcmp("regex", mod->name) ) {
			if ( mfunc==str_match_nocase) {
				mesg(M_ERR,"match_str: nocase and "
					"regex mutually exclusive.");
				free(p);
				return 0;
			}
			mfunc=str_match_regex;
		}else if ( !strcmp("nocase", mod->name) ) {
			if ( mfunc==str_match_regex ) {
				mesg(M_ERR,"match_str: nocase and "
					"regex mutually exclusive.");
				free(p);
				return 0;
			}
			mfunc=str_match_nocase;
		}
	}

	p->flag_uri=mode;

	/* Finished strings can be no bigger than this */
	if ( !(p->string=malloc(j)) ) {
		free(p);
		return 0;
	}

	for(p->len=j,t=args,j=0; *t; t++) {
		if ( state==0 ) { /* Normal string */
			if ( !escaped && *t=='\\' ) {
				escaped=1;
			}else if ( !escaped && *t=='|' ) {
				state=1;
			}else{
				escaped=0;
				if ( mfunc==str_match_nocase ) {
					p->string[j++]=toupper(
						(unsigned char)*t);
				}else{
					p->string[j++]=*t;
				}
			}
		}else if ( state==1 ) { /* Binary data */
			if ( *t=='|' ) {
				state=0;
				continue;
			}

			if ( *t==' ' || *t=='\t' ) {
				continue;
			}

			bin[i++]=*t;

			if ( i==2 ) {
				unsigned int num;
				char thisbyte;
				i=0;

				if ( !hextouint(bin, &num) ) {
					mesg(M_ERR,"match_str: bogus hex char");
					free(p->string);
					free(p);
					return 0;
				}

				/* Impossible to be too big cos
				 * 16^2 = 0xff */
				thisbyte=(char)num&0xff;

				if ( mfunc==str_match_nocase) {
					p->string[j++]=toupper(
						(unsigned char)thisbyte);
				}else{
					p->string[j++]=thisbyte;
				}
			}
		}
	}

	if ( i ) {
		mesg(M_ERR,"match_str: string is n and a half bytes!");
		free(p->string);
		free(p);
		return 0;
	}

	/* Shrink the string if there was any
	 * binary data inside of it */
	if ( j != p->len ) {
		realloc(p->string, j); /* can only be smaller */
		p->len=j;
	}

	/* Length can't be bigger than depth */
	if ( p->depth && p->len > p->depth ) {
		mesg(M_ERR, "match_str: len(%u) > depth(%u)", p->len, p->depth);
		free(p->string);
		free(p);
		return 0;
	}

	p->type = mfunc;

	*priv=p;

	return 1;
}

static int content_validate(struct matcher *match, char *args,
				void **priv, struct modifier *m)
{
	return str_validate(args, priv, m, 0);
}

static int uricontent_validate(struct matcher *match, char *args,
				void **priv, struct modifier *m)
{
	return str_validate(args, priv, m, 1);
}

static void str_cleanup(void *priv) {
	struct content_priv *p=priv;

	if ( p ) {
		if ( p->string )
			free(p->string);
		free(p);
	}
}

static int action(void *priv, int index, void *data)
{
	struct content_priv *p = (struct content_priv *)priv;
	struct packet *pkt = (struct packet *)data;

	if ( p->alert )
		alert_tag(pkt, p->alert, index);

	if ( p->result )
		p->result->match(p->result, pkt, index);

	return 1;
}

static inline void do_search(struct content_detect *x, struct packet *p,
				int l, char *hs, size_t hlen)
{
	struct content_priv *n;

	for(n=x->raw; n; n=n->next) {
		int ret = n->type(n, hs, (int)hlen);
		if ( n->negate ^ ret ) {
			if ( n->alert )
				alert_tag(p, n->alert, l);

			if ( n->result )
				n->result->match(n->result, p, l);
		}
	}

	mwmSearch(x->pat, hs, hlen, action, p, l);
}

static void content_match(struct detect *d, struct packet *p, int l)
{
	struct content_detect *x = (struct content_detect *)d;
	int origl = l;
	size_t hlen;
	char *hs;

	l++; /* ignore IP */
	if ( p->layer[l+1].proto )
		l++; /* ignore transport layer header */
	if ( l >= p->llen )
		goto skip;

	hlen = p->end - p->layer[l].h.raw;
	if ( hlen == 0 )
		goto skip;

	hs = p->layer[l].h.raw;

	do_search(x, p, origl, hs, hlen);

skip:
	if ( x->d.any )
		x->d.any->match(x->d.any, p, origl);
}

static void uricontent_match(struct detect *d, struct packet *p, int l)
{
	struct content_detect *x = (struct content_detect *)d;
	struct http_request *h;
	int origl = l;
	size_t hlen;
	char *hs;

	l++; /* ignore IP */
	if ( p->layer[l+1].proto )
		l++; /* ignore transport layer header */
	if ( l >= p->llen )
		goto skip;

	if ( p->layer[l].proto != http_proto )
		goto skip;
	if ( !(h=p->layer[l].session) )
		goto skip;
	if ( h->type != HTTP_REQUEST )
		goto skip;
	if ( !(hlen=h->uri.len) )
		goto skip;
	if ( !(hs=h->uri.ptr) )
		goto skip;

	do_search(x, p, origl, hs, hlen);

skip:
	if ( x->d.any )
		x->d.any->match(x->d.any, p, origl);
}

static struct detect *shared_content_new(int uri)
{
	struct content_detect *x;

	x = calloc(1, sizeof(*x));
	if ( x == NULL )
		return NULL;

	x->pat = mwmNew();
	if ( x->pat == NULL ) {
		free(x);
		return NULL;
	}

	if ( uri ) {
		x->d.match = uricontent_match;
		mwmLargeShifts(x->pat, 1);
	}else{
		x->d.match = content_match;
	}

	return (struct detect *)x;
}

static struct detect *content_new(struct matcher *m)
{
	return shared_content_new(0);
}

static struct detect *uricontent_new(struct matcher *m)
{
	return shared_content_new(1);
}

static int content_optimize(struct detect *d)
{
	struct content_detect *x = (struct content_detect *)d;
	if ( mwmPrepPatterns(x->pat) )
		return 0;
	return 1;
}

static struct detect **content_insert(struct detect *d, struct sig_node *n)
{
	struct content_detect *x = (struct content_detect *)d;
	struct content_priv *p=(struct content_priv *)n->private;

	p->negate = n->negate;
	p->alert = n->alert;

	if ( p->negate || p->type == str_match_regex ) {
		p->skip = calloc(1, sizeof(*p->skip) * 256);
		if ( p->skip == NULL )
			return NULL;
		bm_skip(p->string, p->len, p->skip);

		p->next = NULL;
		x->raw = p;
		return &p->result;
	}

	mwmAddPatternEx(x->pat,
			p->string,
			p->len,
			p->type == str_match_nocase,
			p->offset,
			p->depth,
			p, x->id++);

	return &p->result;
}

static struct matcher match_content = {
	.validate = content_validate,
	.free_priv = str_cleanup,
	.new_node = content_new,
	.optimize = content_optimize,
	.insert_node = content_insert,
};

static struct matcher match_uricontent = {
	.validate = uricontent_validate,
	.free_priv = str_cleanup,
	.new_node = uricontent_new,
	.optimize = content_optimize,
	.insert_node = content_insert,
};

static struct field builtin_fields[] = {
	{
		.name = "content",
		.desc = "Contains String",
		.type = IDX_INVALID,
		.matcher = &match_content,
		.cost = MCOST_DATA + 1000,
		.allow_opts = 1,
	},
	{
		.name = "uricontent",
		.desc = "URI Contains String",
		.type = IDX_INVALID,
		.matcher = &match_uricontent,
		.cost = MCOST_DATA + 1000,
		.allow_opts = 1,
	},
	{.name = NULL }
};

struct field_reg firestorm_fields[] = {
	{ .field = builtin_fields },
	{ .field = NULL }
};

/* Wu-Mamber setwise string matching */
struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "matcher.str",
	.desc		= "String matching routines",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 2,
	.ver_minor	= 0,
};
