/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
 *
 * snort - Snort rulefile parser
 *
 * Generally the parsing is done with turing state machines. It has become
 * evil and hacky over time, and is in need of some serious tidyups.
 *
 * TODO:
 *  o Cleanup the various hacks with a keyword -> function table
 *  o Full snort variable support
 *  o Multi-line support
*/

#include <firestorm.h>
#include <dotscara/fdctl.h>
#include <dotscara/flbuf.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/unistd.h>
#include <fcntl.h>
#include <ctype.h>

#include <args.h>
#include <strtouint.h>
#include <plugin.h>
#include <alert.h>
#include <signature.h>
#include <parser.h>

/* File state */
struct snort_file {
	const char *fn;
	size_t line;
};

/* Maximum line length */
#define SNORT_MAXLEN 8192

/* Used for rule ordering */
static int nr;

/* Snort variables */
struct snort_var {
	struct snort_var	*next;
	char			*name;
	char			*value;
	int			negate;
};
static struct snort_var *vars;

/* Snort rule classification */
struct snort_class {
	struct snort_class *next;
	char *name;
	u_int8_t prio;
};
static struct snort_class *c_list;

/* Snort keywords */
struct skw{
	const char *name;
	int (*fn)(const char *, size_t, char *);
};

static int snort_line(const char *, size_t, char *);
static int snort_var(const char *, size_t, char *);
static int snort_config(const char *, size_t, char *);

static struct skw snort_keywords[]={
	{"alert", snort_line},
	{"log", snort_line},
#if 0
	{"activate", NULL},
	{"dynamic", NULL},
	{"pass", NULL},
#endif
	{"config", snort_config},
	{"var", snort_var},
	{NULL, NULL}
};

struct tag_token {
	const char *str;
	unsigned int val;
};

static struct tag_token tag_types[]={
	{"session", TAG_TYPE_SESSION},
	{"host", TAG_TYPE_HOST},
	{NULL, }
};

static struct tag_token tag_metrics[]={
	{"packets", TAG_METRIC_PACKETS},
	{"seconds", TAG_METRIC_SECONDS},
	{"bytes", TAG_METRIC_BYTES},
	{NULL, }
};

static struct tag_token tag_dirs[]={
	{"src", TAG_DIR_SRC},
	{"dst", TAG_DIR_DST},
	{"src_only", TAG_DIR_SRC_ONLY},
	{"dst_only", TAG_DIR_DST_ONLY},
	{"server", TAG_DIR_SERVER},
	{"client", TAG_DIR_CLIENT},
	{"both", TAG_DIR_BOTH},
	{"either", TAG_DIR_EITHER},
	{NULL, }
};

/* Snort parsing variables */
#define SNORT_PROTO 0
#define SNORT_SADDR 1
#define SNORT_SPORT 2
#define SNORT_DIR 3
#define SNORT_DADDR 4
#define SNORT_DPORT 5
#define SNORT_MAX 6
static char *snort_head[SNORT_MAX];
static int snort_negate[SNORT_MAX];

static struct criteria *se_head;
static struct criteria *se_tail;
static struct criteria *src_additional;
static struct criteria *dst_additional;

static char *snort_msg;
static unsigned int snort_rate;
static unsigned int snort_burst;
static u_int32_t snort_prio;
static u_int32_t snort_prio_set;
static u_int32_t snort_sid, snort_rev;

static int tag_set;
static struct alert_tag the_tag;

/* These are modifiers */
static char *snort_modifiers[]={
	"nocase",
	"offset",
	"depth",
	"regex",
	NULL
};

/* Not implemented hacks */
static char *snort_hacks[]={
	"reference",
	"logto",
	NULL
};

static int snort_commit_rule(int index);
static int snort_iplist(int index);
static struct {
	int var;
	int(*fn)(int);
}steps[]={
	{SNORT_SADDR, snort_iplist},
	{SNORT_DADDR, snort_iplist},
	{0, snort_commit_rule},
};

/* Error string */
static char estr[512];

static void cleanup_additional(struct criteria *x)
{
	struct criteria *s;
	for(s=x; s; ) {
		struct criteria *tmp=s;
		s=s->next;
		if ( tmp->args )
			free(tmp->args);
		free(tmp);
	}
}

static struct snort_var *snort_var_find(char *key)
{
	struct snort_var *ret;

	for(ret=vars; ret; ret=ret->next) {
		if ( !strcmp(key, ret->name) )
			return ret;
	}

	snprintf(estr, sizeof(estr), "snort var not found: $%s", key);
	return NULL;
}

static struct snort_class *snort_class_find(char *key)
{
	struct snort_class *ret;
	for(ret=c_list; ret; ret=ret->next) {
		if ( !strcmp(key, ret->name) )
			return ret;
	}

	return NULL;
}

static int snort_tag_token(struct tag_token *t,
				const char *str,
				unsigned int *val)
{
	for(; t->str; t++) {
		if ( !strcasecmp(str, t->str) ) {
			*val = t->val;
			return 1;
		}
	}

	return 0;
}

static int snort_parse_tag(char *str)
{
	char *toks[4];
	int num_toks;

	num_toks = easy_explode(str, ',', toks, 4);
	if ( num_toks < 3 )
		return 0;

	the_tag.direction = TAG_DIR_BOTH;

	if ( !snort_tag_token(tag_types, toks[0], &the_tag.type) )
		return 0;
	if ( strtouint(toks[1], &the_tag.count) )
		return 0;
	if ( !snort_tag_token(tag_metrics, toks[2], &the_tag.metric) )
		return 0;
	if ( num_toks >= 4 &&
		!snort_tag_token(tag_dirs, toks[3], &the_tag.direction) )
		return 0;

	tag_set = 1;

	return 1;
}

static int snort_parse_rate(char *str)
{
	unsigned int num;
	timestamp_t d;
	int i;

	if ( (i=strtouint(str,&num))<0 )
		goto err;

	if ( i ) {
		str+=i;
		if ( *str++ != '/' )
			goto err;
		switch ( *str ) {
		case 's':
		case 'S':
			d = RATE_SEC;
			break;
		case 'm':
		case 'M':
			d = RATE_MIN;
			break;
		case 'h':
		case 'H':
			d = RATE_HR;
			break;
		case 'd':
		case 'D':
			d = RATE_DAY;
			break;
		default:
			goto err;
		}
	}else{
		/* Seconds is the default */
		d = RATE_SEC;
	}

	if ( num>d ) {
		snprintf(estr,sizeof(estr),"maximum granularity is 100/sec: %s", str);
		return 0;
	}

	snort_rate=d/num;
	return 1;
err:
	snprintf(estr,sizeof(estr),"rate parse error for '%s'", str);
	return 0;
}

static inline char *strip_escapes(char *val)
{
	char *in, *out;
	int state;

	for(in=out=val, state=0; *in != '\0'; in++) {
		if ( state == 0 ) {
			if ( *in == '\\' ) {
				state = 1;
				continue;
			}
		}else{
			state = 0;
		}
		*out = *in;
		out++;
	}

	*out = '\0';

	return val;
}

static int snort_addcrit(char *key, char *val, int negate)
{
	int i;
	struct criteria *se;

	/* Snatch alert meta-data */
	if ( !strcmp(key, "msg") ) {
		snort_msg = strip_escapes(val);
		return 1;
	}
	if ( !strcmp(key, "sid") ) {
		if ( strtouint(val, &snort_sid) ) {
			snprintf(estr,sizeof(estr),"invalid sid '%s'", val);
			return 0;
		}
		return 1;
	}
	if ( !strcmp(key, "rev") ) {
		if ( strtouint(val, &snort_rev) ) {
			snprintf(estr,sizeof(estr),"invalid rev '%s'", val);
			return 0;
		}
		return 1;
	}
	if ( !strcmp(key, "priority") ) {
		if ( strtouint(val, &snort_prio) ) {
			snprintf(estr,sizeof(estr),"invalid priority '%s'", val);
			return 0;
		}
		if ( snort_prio & ~0xffUL ) {
			snprintf(estr,sizeof(estr),"priority is out of range (0-255)");
			return 0;
		}
		snort_prio_set=1;
		return 1;
	}
	if ( !strcmp(key, "classtype") ) {
		struct snort_class *c;
		if ( !(c=snort_class_find(val)) ) {
			snprintf(estr,sizeof(estr),"Unknown classtype '%s'", val);
			return 0;
		}
		if ( !snort_prio_set ) snort_prio=c->prio;
		return 1;
	}

	/* Rate limtiing */
	if ( !strcmp(key, "rate") ) {
		return snort_parse_rate(val);
	}
	if ( !strcmp(key, "burst") ) {
		if ( strtouint(val, &snort_burst) ) {
			snprintf(estr,sizeof(estr),"invalid burst '%s'", val);
			return 0;
		}
		return 1;
	}

	/* Tagging */
	if ( !strcmp(key, "tag") ) {
		return snort_parse_tag(val);
	}

	for(i=0; snort_hacks[i]; i++) {
		if ( !strcmp(key, snort_hacks[i]) )
			return 1;
	}

	/* Modifier options */
	for(i=0; snort_modifiers[i]; i++) {
		struct modifier *m;

		if ( !strcmp(key, snort_modifiers[i]) ) {
			if ( !se_tail )
				return 0;
			if ( !(m=calloc(1, sizeof(*m))) )
				return 0;
			m->name=key;
			m->args=val;
			m->negate=negate;
			m->next=se_tail->modifier;
			se_tail->modifier=m;
			return 1;
		}
	}

	if ( !(se=calloc(1, sizeof(*se))) ) {
		return 0;
	}

	se->next=NULL;
	se->match=(struct matcher *)key;
	se->args=val;
	se->negate=negate;

	if ( se_tail ) {
		se_tail->next = se;
	}else{
		se_head = se;
	}
	se_tail = se;

	return 1;
}

/* Parse the body of a snort rule (the bit after the bracket) */
#define zero_body() key=NULL; val=NULL; neg=0; state=0;
static int snort_parse_body(char *line)
{
	char *key=NULL;
	char *val=NULL;
	int neg=0;

	char *cur;
	int state=0;
	int inq=0;
	int esc=0;

	/* Snort state machine parser */
	for(cur=line; *cur; cur++) {
		if ( state==0 ) { /* Anywhere */
			if ( !isspace(*cur) ) {
				key=cur;
				state=1;
			}else if ( *cur==')' ) {
				*cur='\0';
				break;
			}
		}else if ( state==1 ) { /* Inside a key */
			if ( *cur==':' ) {
				*cur='\0';
				state=2;
			}else if ( *cur==';' && *(cur-1)!='\\' ) {
				*cur='\0';
				val=NULL;

				if ( !snort_addcrit(key, val, neg) )
					return 0;
				zero_body();
			}
		}else if ( state==2 ) { /* Between key and value */
			if ( *cur=='!' ) {
				neg=1;
			}else if ( !inq && *cur==';' ) {
				state = 3;
				cur--;
			}else if ( !isspace(*cur) ) {
				if ( *cur=='\"' ) {
					inq=1;
				}

				val=cur + inq;
				state=3;
			}
		}else if ( state==3 ) { /* In a value */
			/* Enter quote marks */
			if ( !inq && *cur=='\"' ) {
				inq=1;
				continue;
			}

			/* Allow missing end semicolon */
			if ( !inq && *cur==')' ) {
				if ( *(cur-1)=='\"' )
					*(cur-1)='\0';
				*cur='\0';

				if ( !snort_addcrit(key, val, neg) )
					return 0;
				zero_body();
				continue;
			}

			/* Check for escape chars */
			if ( !esc && *cur=='\\' ) {
				esc=1;
				continue;
			}

			if ( inq && !esc && *cur=='\"' ) {
				/* End quotes */
				inq=0;
			}else if ( !inq && !esc && *cur==';' ) {
				/* End  */
				if ( *(cur-1)=='\"' )
					*(cur-1)='\0';
				*cur='\0';

				if ( !snort_addcrit(key, val, neg) )
					return 0;
				zero_body();
			}else{
				esc=0;
			}
		}
	}

	return 1;
}

/* Actually commit a rule into the signature
 * subsystem */
static int snort_commit_rule(int index)
{
	struct criteria *s;
	struct modifier *m;
	struct rule *r;

	/* Create a new rule */
	if ( !(r=signature_rule(snort_head[SNORT_PROTO])) ) {
		snprintf(estr, sizeof(estr), "Unable to create new '%s' rule",
			snort_head[SNORT_PROTO]);
		return 0;
	}

	/* Alert stuff. Just use default alerting mechanism */
	if ( !snort_sid )
		mesg(M_WARN,"snort: rule has no sid: %s", snort_msg);

	if ( !signature_alert(r, snort_msg, ++nr,
			snort_sid, snort_rev,
			snort_rate, snort_burst,
			(u_int8_t)snort_prio & 0xff) ) {
		snprintf(estr, sizeof(estr), "Unable to set rule target");
		goto abort;
	}

	if ( tag_set && !signature_tag(r, &the_tag) ) {
		snprintf(estr, sizeof(estr), "Unable to set rule tag");
		goto abort;
	}

	/* Source and destination addresses */
	if ( strcmp(snort_head[SNORT_SADDR], "any") &&
		!signature_criteria(r, snort_negate[SNORT_SADDR],
		"src", snort_head[SNORT_SADDR]) ){
		snprintf(estr, sizeof(estr), "Unable to add head criteria to rule");
		goto abort;
	}
	if ( strcmp(snort_head[SNORT_DADDR], "any") &&
		!signature_criteria(r, snort_negate[SNORT_DADDR],
		"dst", snort_head[SNORT_DADDR]) ){
		snprintf(estr, sizeof(estr), "Unable to add head criteria to rule");
		goto abort;
	}


	/* Ports */
	if ( strcmp(snort_head[SNORT_SPORT], "any") ) {
		if ( !signature_criteria(r, snort_negate[SNORT_SPORT],
				"sport", snort_head[SNORT_SPORT]) ) {
			snprintf(estr, sizeof(estr), "Unable to add head criteria to rule");
			goto abort;
		}
	}
	if ( strcmp(snort_head[SNORT_DPORT], "any") ) {
		if ( !signature_criteria(r, snort_negate[SNORT_DPORT],
				"dport", snort_head[SNORT_DPORT]) ) {
			snprintf(estr, sizeof(estr), "Unable to add head criteria to rule");
			goto abort;
		}
	}

	/* Additional source and destination addresses */
	for(s=src_additional; s; s=s->next) {
		if ( !signature_criteria(r, s->negate, "src", s->args) ){
			snprintf(estr, sizeof(estr), "Unable to add head criteria to rule");
			goto abort;
		}
	}
	for(s=dst_additional; s; s=s->next) {
		if ( !signature_criteria(r, s->negate, "dst", s->args) ){
			snprintf(estr, sizeof(estr), "Unable to add head criteria to rule");
			goto abort;
		}
	}

	/* And all the options... */
	for(s=se_head; s; s=s->next) {
		if ( !signature_criteria(r, s->negate,
					(char *)s->match, s->args) ) {
			snprintf(estr, sizeof(estr),
					"Unable to add '%s' criteria to rule",
					(char *)s->match);
			goto abort;
		}
		for(m=s->modifier; m; m=m->next) {
			if ( !signature_modifier(r, m->negate,
					m->name, m->args) ) {
				snprintf(estr, sizeof(estr),
					"Unable to add '%s' modifier to rule",
					m->name);
				goto abort;
			}
		}
	}

	/* lets go */
	if ( !signature_commit(r) ) {
		snprintf(estr, sizeof(estr), "Rule failed to commit: %s",
			snort_msg);
		goto abort;
	}

	return 1;

abort:
	signature_abort(r);
	return 0;
}

static int snort_iplist(int index)
{
	char buf[128];
	char *cur, *ip, *tmp;
	size_t len=1;
	int ret=1;
	int i=steps[index++].var;
	struct criteria **additional;

	if ( snort_head[i][0]!='[' )
		return steps[index].fn(index);

	if ( i==SNORT_SADDR ) {
		additional=&src_additional;
	}else if ( i==SNORT_DADDR ) {
		additional=&dst_additional;
	}else return 0;

	*additional=NULL;

	tmp=snort_head[i];
	snort_head[i]=buf;
	for(ip=cur=tmp+1; ; cur++, len++) {
		if ( *cur==',' || *cur==']' || *cur==0 ) {
			if ( len > sizeof(buf) ) {
				mesg(M_WARN,"snort: ip in list too long!");
				break;
			}

			memcpy(buf, ip, len-1);
			buf[len-1]=0;

			if ( snort_negate[i] && *cur==',' ) {
				struct criteria *foo;
				if ( !(foo=calloc(1, sizeof(*foo))) ) {
					ret=0;
					break;
				}

				if ( !(foo->args=strdup(buf)) ) {
					free(foo);
					ret=0;
					break;
				}
				foo->match=NULL;
				foo->negate=1;
				foo->next=*additional;
				*additional=foo;
			}else{
				if ( !steps[index].fn(index) ) {
					ret=0;
					break;
				}
			}

			if ( *cur==',') {
				ip=cur+1;
				len=0;
				continue;
			}else break;
		}
	}

	if ( *additional ) {
		cleanup_additional(*additional);
		*additional=NULL;
	}

	snort_head[i]=tmp;
	return ret;
}

/* See if a string is a variable if so, swap it */
static int snort_var_replace(char **ptr, int *n)
{
	struct snort_var *v;

	if ( **ptr=='$' ) {
		if ( !(v=snort_var_find(*ptr+1)) ) {
			return 0;
		}
		*ptr=v->value;
		*n=v->negate;
		return 1;
	}else return 1;
}

static int add_classification(char *name, char *prio)
{
	unsigned int val;
	struct snort_class *c;

	if ( strtouint(prio, &val) ) {
		snprintf(estr,sizeof(estr),"priorities must be numeric");
		return 0;
	}

	if ( val & ~0xffUL ) {
		snprintf(estr,sizeof(estr),"priority is out of range (0-255)");
		return 0;
	}

	if ( (c=snort_class_find(name)) ) {
		snprintf(estr,sizeof(estr),"re-definition of '%s' classification", name);
		return 0;
	}

	if ( !(c=calloc(1, sizeof(*c))) ) {
		snprintf(estr,sizeof(estr),"calloc: %s", sys_err());
		return 0;
	}

	if ( !(c->name=strdup(name)) ) {
		snprintf(estr,sizeof(estr),"strdup: %s", sys_err());
		free(c);
		return 0;
	}

	c->prio=(u_int8_t)val&0xff;

	c->next=c_list;
	c_list=c;

	return 1;
}

/* Deal with config lines in snort files */
static int snort_config(const char *fn, size_t ln, char *line)
{
	char *cur;
	char *str[3];
	int i=0;
	int state=0;

	for(cur=line; *cur; cur++) {
		if ( *cur==':' ) {
			*cur=0;
			cur++;
			break;
		}
	}

	if ( strcmp(line, "classification") ) {
		mesg(M_WARN,"snort: %s:%u: ignoring '%s' config",
			fn, ln, line);
		return 1;
	}

	str[0]=cur;
	str[1]=NULL;
	str[2]=NULL;
	for(; *cur; cur++) {
		if ( i>2 ) {
			continue;
		}else if ( state==0 ) {
			if ( !isspace(*cur) ) {
				str[i++]=cur;
				state=1;
			}
		}else if ( state==1 ) {
			if ( *cur==',' ) {
				*cur=0;
				state=0;
			}
		}
	}

	if ( i<3 ) {
		snprintf(estr,sizeof(estr), "malformed classification");
		return 0;
	}

	return add_classification(str[0],str[2]);
}

/* Parse a 'var' line */
static int snort_var(const char *fn, size_t ln, char *line)
{
	char *key=NULL, *val=NULL;
	char *cur;
	struct snort_var *vn;
	int end=0;
	int negate=0, nneg=0;
	int state=0;

	for(cur=line; *cur; cur++) {
		if ( state==0 ) {
			if ( *cur=='!' && key ) {
				negate=1;
			}else if ( !isspace(*cur) ) {
				if ( !key ) {
					key=cur;
					state=1;
				}else{
					if ( !val ) val=cur;
					continue;
				}
			}
		}else if ( state==1 ) {
			if ( isspace(*cur) ) {
				*cur='\0';
				state=0;
				end=1;
			}
		}
	}

	if ( !end ) {
		snprintf(estr, sizeof(estr), "Incomplete line");
		return 0;
	}

	/* TODO: This should be redundant */
	if ( !snort_var_replace(&val, &nneg) ) {
		negate = negate ^ nneg;
		return 0;
	}

	if ( (vn=snort_var_find(key)) ) {
		char *snew;

		if ( !(snew=strdup(val)) ) {
			goto err_nomem;
		}

		free(vn->value);
		vn->value=snew;
		vn->negate=negate;
	}else{
		if ( !(vn=malloc(sizeof(*vn))) ) {
			goto err_nomem;
			return 0;
		}

		if ( !(vn->name=strdup(key)) ) {
			free(vn);
			goto err_nomem;
		}

		if ( !(vn->value=strdup(val)) ) {
			free(vn->name);
			free(vn);
			goto err_nomem;
		}

		vn->negate=negate;
		vn->next=vars;
		vars=vn;
	}

	return 1;
err_nomem:
	snprintf(estr, sizeof(estr), "Out of memory for var '%s'", key);
	return 0;
}

/* Snort up each line of the config  */
static int snort_line(const char *fn, size_t ln, char *line)
{
	struct criteria *s;
	struct modifier *m;
	char *cur;
	int i;
	int end=0;
	int state=0;

	/* Zero all global values */
	estr[0]=0;
	tag_set=0;
	snort_sid=0;
	snort_rev=0;
	snort_prio=0;
	snort_prio_set=0;
	for(i=0; i<SNORT_MAX; i++) {
		snort_negate[i]=0;
	}

	/* State machine parser, parse out first 7 elements
	 * seperated by any amount of whitespace ending in a '(' */
	for(i=0, cur=line; *cur; cur++) {
		if ( state == 0 ) {
			if ( *cur=='(' ) {
				if ( i==SNORT_MAX ) end=1;
				break;
			}

			if ( *cur=='!' ) {
				snort_negate[i]=1;
			}else if ( !isspace(*cur) && *cur ) {
				if ( i >= SNORT_MAX ) break;

				snort_head[i]=cur;
				state=1;
				i++;
			}
		}else if ( state == 1 ) {
			if ( isspace(*cur) ) {
				*cur='\0';
				state=0;
			}
		}
	}

	/* Parse error */
	if ( !end ) {
		snprintf(estr, sizeof(estr), "Incomplete line");
		return 0;
	}

	/* Its either "<-", "->" or "<>" */
	if ( strlen(snort_head[SNORT_DIR]) !=2 ) {
		snprintf(estr, sizeof(estr), "Invalid direction '%s'",
			snort_head[SNORT_DIR]);
		return 0;
	}

	/* Parse the rest of the rule */
	se_head=NULL;
	se_tail=NULL;
	snort_msg=NULL;
	snort_rate=0;
	snort_burst=0;
	if ( !snort_parse_body(++cur) ) return 0;

	/* Do variable substitutions */
	if ( !snort_var_replace(&snort_head[SNORT_SADDR],
		&snort_negate[SNORT_SADDR]) ) return 0;
	if ( !snort_var_replace(&snort_head[SNORT_SPORT],
		&snort_negate[SNORT_SPORT]) ) return 0;
	if ( !snort_var_replace(&snort_head[SNORT_DADDR],
		&snort_negate[SNORT_DADDR]) ) return 0;
	if ( !snort_var_replace(&snort_head[SNORT_DPORT],
		&snort_negate[SNORT_DPORT]) ) return 0;

	/* Go forwards */
	if ( snort_head[SNORT_DIR][1]=='>' ) {
		if ( !steps[0].fn(0) ) {
			return 0;
		}
	}

	/* Go backwards */
	if ( snort_head[SNORT_DIR][0]=='<' ) {
		char *tadr, *tprt;
		int tna, tnp;

		/* Swap IPs and ports */
		tadr=snort_head[SNORT_SADDR];
		snort_head[SNORT_SADDR]=snort_head[SNORT_DADDR];
		snort_head[SNORT_DADDR]=tadr;

		tna=snort_negate[SNORT_SADDR];
		snort_negate[SNORT_SADDR]=snort_negate[SNORT_DADDR];
		snort_negate[SNORT_SPORT]=tna;

		tprt=snort_head[SNORT_SPORT];
		snort_head[SNORT_SPORT]=snort_head[SNORT_DPORT];
		snort_head[SNORT_DPORT]=tprt;

		tnp=snort_negate[SNORT_SPORT];
		snort_negate[SNORT_DADDR]=tna;
		snort_negate[SNORT_SPORT]=snort_negate[SNORT_DPORT];

		if ( !steps[0].fn(0) ) {
			return 0;
		}
	}

	/* Free everything */
	for(s=se_head; s; ) {
		struct criteria *tmp=s;
		for(m=s->modifier; m; ) {
			struct modifier *tmp=m;
			m=m->next;
			free(tmp);
		}
		s=s->next;
		free(tmp);
	}

	return 1;
}

/* Look at the first word of a snort line and decide
 * what kind of line it is and how to deal with it
*/
static int snort_dispatch(const char *fn, size_t ln, char *line)
{
	struct skw *k;
	char *cur, *kw=NULL;
	int state=0;
	int end=0;

	for(cur=line; *cur; cur++) {
		if ( state==0 ) {
			if ( !isspace(*cur) ) {
				if ( !kw ) {
					kw=cur;
					state=1;
				}else{
					break;
				}
			}
		}else if ( state==1 ) {
			if ( isspace(*cur) ) {
				*cur='\0';
				state=0;
				end=1;
			}
		}
	}

	if ( !end )
		return 0;

	for(k=snort_keywords; k->name; k++) {
		if ( !strcmp(kw, k->name) )
			return k->fn(fn, ln, cur);
	}

	snprintf(estr, sizeof(estr), "Unsupported snort keyword: '%s'", kw);
	return 0;
}

static int cbfn(struct snort_file *f, u_int8_t *buf, size_t line, size_t len)
{
	f->line = line;

	while( *buf && isspace(*buf) ) {
		buf++;
		len--;
	}

	if ( *buf == '#' || *buf == '\0' )
		return 1;

	return snort_dispatch(f->fn, f->line, buf);
}

static int cbfn2(struct snort_file *f, u_int8_t *buf, size_t line, size_t len)
{
	f->line = line;
	snprintf(estr, sizeof(estr), "Line too long");
	return 0;
}

/* Parse the snort file, line by line */
static int snort_run(const char *fn)
{
	struct snort_file f;
	struct flbuf b;
	int fd;

	estr[0] = '\0';

	if ( !fn )
		return 0;

	fd = open(fn, O_RDONLY);
	if ( fd < 0 ) {
		mesg(M_ERR,"snort: %s: open(): %s",
			fn, sys_err());
		return 0;
	}

	f.fn = fn;
	f.line = 0;

	if ( !flbuf_attach(&b, fd, 1024, (flbuf_cbfn_t)cbfn,
				(flbuf_cbfn_t)cbfn2, &f) ) {
		mesg(M_ERR,"snort: %s: flbuf_attach(): %s",
				fn, sys_err2(NULL));
		fd_close(fd);
		return 0;
	}

	if ( !flbuf_read(&b) ) {
		mesg(M_ERR,"snort: %s:%u: %s", fn, f.line, sys_err2(estr));
		fd_close(fd);
		return 0;
	}

	flbuf_close(&b);
	return 1;
}

struct parser firestorm_parsers[]={
	{ .name="snort", .load_file=snort_run},
	{.name = NULL }
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "parser.snort",
	.desc		= "Snort ruleset loader",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 2,
	.ver_minor	= 0,
};
