/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * This file reads in the configuration file and sets up all of
 * the enviroment for the firestorm NIDS sensor to run. Default
 * settings try to be as secure as possible, we are quite strict
 * if you want to configure firestorm in an obviously insecure
 * way, you can recompile with checks removed.
 *
 * TODO
 *  o preserve filename/line number for error messages
 *  o allocation list items as 1 allocation
 *  o tidy up repeated code
 */
#include <firestorm.h>
#include <dotscara/fdctl.h>
#include <dotscara/flbuf.h>

#include <stdarg.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sched.h>
#include <fcntl.h>

#include <args.h>
#include <loader.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>
#include <capdev.h>
#include <exporter.h>
#include <parser.h>
#include <idx.h>
#include <field.h>
#include <nbio.h>
#include <sensor.h>

#ifndef _BSD_SOURCE
#define _BSD_SOURCE
#endif
#include <grp.h>

#if HAVE_PWD_H
#include <pwd.h>
#endif

/* limited by AVAL_SIZE in most cases anyway */
#define CONF_MAXLEN 4096

struct conf {
	struct conf *next;
	char *data;
};

/* godawful */
static struct conf *cnf_plugins, *cnf_plugins_tail;
static struct conf *cnf_plugin, *cnf_plugin_tail;
static struct conf *cnf_preproc, *cnf_preproc_tail;
static struct conf *cnf_sig, *cnf_sig_tail;
static struct conf *cnf_out, *cnf_out_tail;
static struct conf *cnf_cap, *cnf_cap_tail;

static unsigned int f_uid;
static unsigned int f_gid;
static unsigned int do_chroot = 1;
static unsigned int realtime = 1;
static char *rootfn, *logfile;

static struct passwd *p;

/* Callbacks that can only be called once */
static int conf_user(struct arg *a, struct fvec *v, void *u)
{
	char *str;

	str = arg_get_string(v);
	if ( str == 0 )
		return 0;

	p = getpwnam(str);
	if ( p == NULL )
		return 0;

	f_uid = p->pw_uid;
	f_gid = p->pw_gid;

	return 1;
}

static int conf_rootfn(struct arg *a, struct fvec *v, void *u)
{
	char *str;

	if ( rootfn )
		return 0;

	str = arg_get_string(v);
	if ( str == 0 )
		return 0;

	rootfn = strdup(str);
	if ( rootfn == NULL )
		return 0;

	return 1;
}

static int conf_logfile(struct arg *a, struct fvec *v, void *u)
{
	char *str;

	if ( logfile )
		return 0;

	str = arg_get_string(v);
	if ( str == 0 )
		return 0;

	logfile = strdup(str);
	if ( logfile == NULL )
		return 0;

	return 1;
}

/* Add a line to a list of lines */
/* XXX: Allocate as 1 chunk */
static int conf_addline(struct conf **head, struct conf **tail, struct fvec *v)
{
	struct conf *c;
	char *str;

	str = arg_get_string(v);
	if ( str == 0 )
		return 0;

	if ( !(c=calloc(1, sizeof(*c))) )
		firestorm_die_error("calloc");

	if ( !(c->data=strdup(str)) )
		firestorm_die_error("strdup");

	if ( !(*head) )  {
		(*head)=c;
		(*tail)=c;
	}else{
		(*tail)->next=c;
		(*tail)=c;
	}

	return 1;
}

/* Callbacks that can be called multiple times */
static int conf_plugins(struct arg *a, struct fvec *v, void *u)
{
	return conf_addline(&cnf_plugins, &cnf_plugins_tail, v);
}
static int conf_plugin(struct arg *a, struct fvec *v, void *u)
{
	return conf_addline(&cnf_plugin, &cnf_plugin_tail, v);
}
static int conf_preproc(struct arg *a, struct fvec *v, void *u)
{
	return conf_addline(&cnf_preproc, &cnf_preproc_tail, v);
}
static int conf_sig(struct arg *a, struct fvec *v, void *u)
{
	return conf_addline(&cnf_sig, &cnf_sig_tail, v);
}
static int conf_outputfn(struct arg *a, struct fvec *v, void *u)
{
	return conf_addline(&cnf_out, &cnf_out_tail, v);
}
static int conf_capfn(struct arg *a, struct fvec *v, void *u)
{
	return conf_addline(&cnf_cap, &cnf_cap_tail, v);
}

/* Each line in the config is treated as a args key/value pair */
static struct arg firestorm_args[]={
	{"effective_uid", ARGTYPE_PUINT, {.vp_uint = &f_uid}},
	{"effective_gid", ARGTYPE_PUINT, {.vp_uint = &f_gid}},
	{"effective_user", ARGTYPE_CUSTOM, {.v_custom = conf_user}},
	{"chroot", ARGTYPE_PBOOL, {.vp_bool = &do_chroot}},
	{"realtime", ARGTYPE_PBOOL, {.vp_uint = &realtime}},
	{"firestorm_root", ARGTYPE_CUSTOM, {.v_custom = conf_rootfn}},
	{"logfile", ARGTYPE_CUSTOM, {.v_custom = conf_logfile}},
	{"capture", ARGTYPE_CUSTOM, {.v_custom = conf_capfn}},
	{"output", ARGTYPE_CUSTOM, {.v_custom = conf_outputfn}},
	{"load_plugins", ARGTYPE_CUSTOM, {.v_custom = conf_plugins}},
	{"load_plugin", ARGTYPE_CUSTOM, {.v_custom = conf_plugin}},
	{"preprocessor", ARGTYPE_CUSTOM, {.v_custom = conf_preproc}},
	{"signatures", ARGTYPE_CUSTOM, {.v_custom = conf_sig}},
	{NULL,}
};

/* Perform initial file descriptor tweaks */
static void conf_tweak_fds(void)
{
	int fds=sysconf(_SC_OPEN_MAX);

	while(fds>2)
		fd_close(fds--);

	if ( (fds=open("/dev/null", O_RDWR))<0 ) {
		firestorm_exit(EXIT_FAILURE, "/dev/null: open(): %s", sys_err());
	}

	/* Make stdin and stderr be /dev/null */
	if ( fds!=0 && dup2(fds,STDIN_FILENO)<0 )
		firestorm_exit(EXIT_FAILURE, "stdin: dup2(): %s", sys_err());
	if ( fds!=2 && dup2(fds,STDERR_FILENO)<0 )
		firestorm_exit(EXIT_FAILURE, "stderr: dup2(): %s", sys_err());

	if ( fds>2 )
		fd_close(fds);
}

/* Open the config file, we need to do some funky stuff */
static int conf_open(const char *fn)
{
	int fd;

	if ( (fd=open(fn, O_RDONLY))<0 ) {
		firestorm_exit(EXIT_FAILURE, "%s: open(): %s", fn, sys_err());
	}

	/* If we were opened with stdout closed the config
	 * could end up as fd1, that kinda sucks because when
	 * we close it, some other bit of code may open fd1 as
	 * writable! printing to stdout would then corrupt the
	 * file - I know we are never setuid but better to be safe */
	if ( fd == 1 ) {
		/* We know fd3 is closed already */
		if ( dup2(fd, 3)<0 )
			firestorm_exit(EXIT_FAILURE,
					"%s: dup2(): %s", fn, sys_err());

		/* We know stdin is /dev/null, make it stdout too */
		if ( dup2(0, fd)<0 )
			firestorm_exit(EXIT_FAILURE,
					"stdout: dup2(): %s", sys_err());

		fd = 3;
	}

	return fd;
}

/* Go in to the background (properly damnit) */
static void conf_background(void)
{
	switch(fork()) {
		case 0: break;
		case -1:
			firestorm_exit(EXIT_FAILURE, "fork(): %s", sys_err());
		default: _exit(0);
	}

	/* This is so that we can't re-aquire a controlling terminal */
	if (setsid()<0) {
		firestorm_exit(EXIT_FAILURE, "setsid(): %s", sys_err());
	}

	switch(fork()) {
		case 0: break;
		case -1:
			firestorm_exit(EXIT_FAILURE, "fork(): %s", sys_err());
		default: _exit(0);
	}
}

/* Parse a line */
static int cbfn(const char *fn, u_int8_t *buf, size_t line, size_t llen)
{
	char *cur;
	int state=0;
	char *val=NULL;

	while ( *buf && isspace(*buf) ) {
		buf++;
		llen--;
	}

	if ( *buf == '#' )
		return 1;

	for(cur=buf; *cur; cur++) {
		switch(state) {
		case 2:
			continue;
		case 0:
			if ( isspace(*cur) ) {
				state = 1;
				*cur = '\0';
			}
			break;
		case 1:
			if ( !isspace(*cur) ) {
				val = cur;
				state = 2;
			}
			break;
		}
	}

	switch ( arg_dispatch(firestorm_args, buf, val, NULL) ) {
	case -1:
		mesg(M_ERR, "%s:%u: parse error: %s", fn, line, buf);
	case 0: /* fall through */
		return 0;
	default:
		return 1;
	}
}

static int cbfn2(const char *fn, u_int8_t *buf, size_t line, size_t llen)
{
	mesg(M_WARN, "%s:%u: Configuration line too long", fn, line);
	return 1;
}

/* Parse the whole file */
static void conf_read(const char *fn, int fd)
{
	struct flbuf b;

	if ( !flbuf_attach(&b, fd,
			CONF_MAXLEN,
			(flbuf_cbfn_t)cbfn,
			(flbuf_cbfn_t)cbfn2,
			(void *)fn) ) {
		firestorm_exit(EXIT_FAILURE, "%s: flbuf_attach: %s",
				fn, sys_err2(NULL));
	}

	if ( !flbuf_read(&b) ) {
		firestorm_exit(EXIT_FAILURE, "%s: %s", fn, sys_err2(NULL));
	}

	flbuf_close(&b);
}

/* parse a preproc line and add it */
static void conf_do_preproc(struct pipeline *p, char *d)
{
	char *tok[2] = {NULL,NULL};
	easy_explode(d, '\0', tok, 2);
	if ( !pipeline_preproc_add(p, tok[0], tok[1]) ) {
		firestorm_exit(EXIT_FAILURE, "couldn't add preproc: %s", tok[0]);
	}
}

/* parse a signature line and add it */
static void conf_do_sig(char *d)
{
	char *tok[2] = {NULL,NULL};
	easy_explode(d, '\0', tok, 2);
	if ( !parser_setup(tok[0], tok[1]) ) {
		firestorm_exit(EXIT_FAILURE, "%s: error loading %s signatures",
				tok[0], tok[1]);
	}
}

/* parse the capture line, and set it all up */
static void conf_capture(struct pipeline *p, char *d)
{
	char *tok[2] = {NULL,NULL};
	easy_explode(d, '\0', tok, 2);

	if ( !pipeline_source_add(p, tok[0], tok[1]) ) {
		firestorm_exit(EXIT_FAILURE, "error starting capture: %s %s",
				tok[0], tok[1]);
	}
}

/* Perform firestorm sensor configuration */
void conf_go(struct pipeline *p, const char *fn)
{
	struct conf *c;
	uid_t init_uid=geteuid(); /* cant fail */

	/* If you want to do something this stupid then you
	 * will have to recompile the code - i wont help you
	 */
	if ( init_uid != getuid() )
		firestorm_exit(EXIT_FAILURE, "SUID is not supported");

	conf_tweak_fds();
	conf_read(fn, conf_open(fn));

	/* Check there actually was even a capture line in the config */
	if ( !cnf_cap )
		firestorm_exit(EXIT_FAILURE, "no captures specified!");

	/* Change to root directory */
	if ( rootfn ) {
	       	if ( chdir(rootfn) )
			firestorm_exit(EXIT_FAILURE, "%s: chdir(): %s",
				rootfn, sys_err());
		free(rootfn);
	}else{
		/* by default change to root, that way firestorm
		 * wont stop the administrator unmounting whatever
		 * filesystem we were run inside of */
		if ( chdir("/") )
			firestorm_exit(EXIT_FAILURE, "/: chdir(): %s", sys_err());
	}

	/* load plugins from directories */
	while(cnf_plugins) {
		c=cnf_plugins;
		cnf_plugins=c->next;
		loader_load_dir(c->data);
		free(c->data);
		free(c);
	}

	/* load individual plugin files */
	while(cnf_plugin) {
		c=cnf_plugin;
		cnf_plugin=c->next;
		if ( !loader_load(c->data) ) {
			firestorm_exit(EXIT_FAILURE, "%s: unable to load plugin", c->data);
		}
		free(c->data);
		free(c);
	}

	/* Load in the actual objects that we want */
	decode_load();
	field_load();
	eventloop_load();
	capdev_load();
	preproc_load();
	parser_load();
	exporter_load();

	if ( !pipeline_init(p, PIPELINE_ALLOW_LIVE|
				PIPELINE_ALLOW_FILE|
				PIPELINE_MERGE) )
		firestorm_exit(EXIT_FAILURE, "pipeline_init failed");

	/* Initialise capture */
	while(cnf_cap) {
		c=cnf_cap;
		cnf_cap=c->next;
		conf_capture(p, c->data);
		free(c->data);
		free(c);
	}

	/* chroot (properly damnit) */
	if ( do_chroot && init_uid ) {
		mesg(M_WARN, "cannot chroot, no privileges");
	}else if ( do_chroot && chroot(".") ) {
		firestorm_exit(EXIT_FAILURE, "chroot(): %s", sys_err());
	}

	/* Drop privileges (properly damnit) */
	if ( !init_uid ) {
		if ( !f_gid || !f_uid )
			firestorm_exit(EXIT_FAILURE,
				"running as root is seriously stupid");

		if ( realtime ) {
			struct sched_param sp;

			sp.sched_priority = sched_get_priority_max(SCHED_RR);

#if _POSIX_MEMLOCK
			if ( mlockall(MCL_FUTURE) )
				firestorm_exit(EXIT_FAILURE,
					"Locking memory: %s", sys_err());
#else
#warning		"POSIX.1b mlockall(2) not available"
#endif

			if ( sched_setscheduler(0, SCHED_RR, &sp) )
				firestorm_exit(EXIT_FAILURE,
					"Setting realtime: %s", sys_err());

			mesg(M_INFO, "Realtime Mode OK (prio=%i)",
					sp.sched_priority);
		}

		if ( setgroups(1,&f_gid) || setgid(f_gid) ) {
			firestorm_exit(EXIT_FAILURE, "unable to drop gid to %u: %s",
				f_gid, sys_err());
		}

		if ( setuid(f_uid) ) {
			firestorm_exit(EXIT_FAILURE, "unable to drop uid to %u: %s",
				f_uid, sys_err());
		}
	}

	/* sort out mesg() output */
	smesg_logfile(logfile);

	/* initialise preprocessors */
	while(cnf_preproc) {
		c=cnf_preproc;
		cnf_preproc=c->next;
		conf_do_preproc(p, c->data);
		free(c->data);
		free(c);
	}

	/* load signatures */
	while(cnf_sig) {
		c=cnf_sig;
		cnf_sig=c->next;
		conf_do_sig(c->data);
		free(c->data);
		free(c);
	}

	if ( logfile ) {
		/* Daemonize (properly damnit) */
		conf_background();
		free(logfile);
	}

	/* Setup exporters */
	while(cnf_out) {
		char *tok[2] = {NULL,NULL};
		c=cnf_out;
		cnf_out=c->next;

		easy_explode(c->data, '\0', tok, 2);
		if ( !pipeline_sink_add(p, SINK_ALERT, tok[0], tok[1]) ) {
			firestorm_exit(EXIT_FAILURE, "output failed to "
					"start: %s %s", tok[0], tok[1]);
		}

		free(c->data);
		free(c);
	}
}
