/*
* This file is part of dotscara
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/poll.h>

#include <dotscara/dotscara.h>
#include <dotscara/sys.h>
#include <dotscara/fdctl.h>
#include <dotscara/fobuf.h>

static int __fobuf_flush(struct fobuf *b) __nonull(1);
static int fobuf_write_slow(struct fobuf *b, const void *buf, size_t len)
	__nonull(2,2);
int inline fobuf_write(struct fobuf *b, const void *buf, size_t len)
	__nonull(1,2);

int inline fobuf_write(struct fobuf *b, const void *buf, size_t len)
{
	if ( likely(len < b->buf_len) ) {
		memcpy(b->ptr, buf, len);
		b->ptr += len;
		b->buf_len -= len;
		return 1;
	}

	return fobuf_write_slow(b, buf, len);
}

static int native_write32(struct fobuf *b, u_int32_t val)
{
	return fobuf_write(b, &val, sizeof(val));
}

static int native_write16(struct fobuf *b, u_int16_t val)
{
	return fobuf_write(b, &val, sizeof(val));
}

static int swap_write32(struct fobuf *b, u_int32_t val)
{
	val = sys_bswap32(val);
	return fobuf_write(b, &val, sizeof(val));
}

static int swap_write16(struct fobuf *b, u_int16_t val)
{
	val = sys_bswap16(val);
	return fobuf_write(b, &val, sizeof(val));
}

static const struct fobuf_ops ops_native = {
	.write32 = native_write32,
	.write16 = native_write16,
};

static const struct fobuf_ops ops_swap = {
	.write32 = swap_write32,
	.write16 = swap_write16,
};

/** __fobuf_flush
 * @param b: the fobuf structure to flush
 *
 * Flush the userspace buffer to disk. Note this does not call fsync()
 * so do not rely on it in order to verify that data is written to disk.
 *
 * Failure modes:
 *  0. return 1: success, all buffered data was written to the kernel
 *  1. undefined: any of the failure modes of fd_write()
 *  2. sig11|sig6|file-corruption: b->buf_len > b->buf_sz
 */
static int __fobuf_flush(struct fobuf *b)
{
	size_t len = b->buf_sz - b->buf_len;
	const void *buf = b->buf;

	/* buffer empty */
	if ( len == 0 )
		return 1;

	if ( !__fd_write(b->fd, buf, len) )
		return 0;

	b->ptr = b->buf;
	b->buf_len = b->buf_sz;

	return 1;
}

int fobuf_flush(struct fobuf *b)
{
	return __fobuf_flush(b);
}

/** fobuf_attach
 * @param b: a fobuf structure to use
 * @param fd: file descriptor to write to
 * @param endian: endianness of the file
 * @param bufsz: size of output buffer, 0 is default
 *
 * Attach a file descriptor to a buffer and ready it for use.
 *
 * Failure modes:
 *  0. return 1: success, buffer was allocated, @b is safe to use.
 *  1. return 0: bufsz > SSIZE_MAX
 *  2. return 0: endianness not defined or supported
 *  3. undefined: any of the failure modes of malloc(3)
 */
int __public fobuf_attach(struct fobuf *b, int fd, int endian, size_t bufsz)
{
	const struct fobuf_ops *ops;

	if ( !bufsz )
		bufsz = 4096;

	if ( bufsz > SSIZE_MAX )
		return 0;

	switch(endian) {
	case DSE_BIG_ENDIAN:
#if __BYTE_ORDER == __BIG_ENDIAN
		ops = &ops_native;
#else
		ops = &ops_swap;
#endif
		break;
	case DSE_LITTLE_ENDIAN:
#if __BYTE_ORDER == __LITTLE_ENDIAN
		ops = &ops_native;
#else
		ops = &ops_swap;
#endif
		break;
	case DSE_NATIVE_ENDIAN:
		ops = &ops_native;
		break;
	default:
		return 0;
	}

	b->buf = malloc(bufsz);
	if ( b->buf == 0 )
		return 0;

	b->fd = fd;
	b->ptr = b->buf;
	b->buf_len = bufsz;
	b->buf_sz = bufsz;
	b->ops = ops;

	return 1;
}

/** fobuf_detach
 * @param b: the fobuf structure to finish up with.
 *
 * Flush the buffer and free up any allocated memory. Note that this
 * does not close the file descriptor, you must do that yourself. It
 * is recommended instead that you use fobuf_close to avoid losing
 * data.
 *
 * Failure modes:
 *  0. return 1: success, buffer was freed, @b is no longer safe
 *  1. undefined: any of the failure modes of __fobuf_flush()
 *  2. undefined: any of the failure modes of free()
 */
int __public fobuf_detach(struct fobuf *b)
{
	int ret = 1;

	if ( !__fobuf_flush(b) )
		ret = 0;

	if ( b->buf )
		free(b->buf);

	b->buf = NULL;
	b->ptr = NULL;
	b->buf_len = 0;

	return ret;
}

/** fobuf_close
 * @param b: the fobuf structure to finish up with.
 *
 * Like fobuf_detach except that we close the file descriptor for
 * writing and we make sure that the data is written safely to disk.
 * before returning.
 *
 * Return value: zero on error, non-zero on success.
 */
int __public fobuf_close(struct fobuf *b)
{
	int ret=1;

	if ( b->fd == -1 )
		goto noclose;

	if ( !__fobuf_flush(b) )
		ret = 0;

	/* don't error if the output file is a special file
	 * which does not support fsync (eg: a pipe)
	 */
	if ( fsync(b->fd) && errno != EROFS && errno != EINVAL )
		ret = 0;

	if ( !__fd_close(b->fd) )
		ret = 0;

noclose:
	if ( b->buf )
		free(b->buf);
	b->buf = NULL;
	b->ptr = NULL;
	b->buf_len = 0;

	return ret;
}

/** fobuf_newfd
 * @param b: the fobuf structure to modify
 * @param fd: the new file descriptor
 *
 * Flush current file safely to disk and start writing to a new FD.
 *
 * Return value: zero on error, non-zero on success.
 */
int __public fobuf_newfd(struct fobuf *b, int fd)
{
	int ret=1;

	if ( b->fd == -1 )
		goto noflush;

	if ( !__fobuf_flush(b) )
		ret = 0;

	/* don't error if the output file is a special file
	 * which does not support fsync (eg: a pipe)
	 */
	if ( fsync(b->fd) && errno!=EROFS && errno!=EINVAL )
		ret = 0;

	if ( !__fd_close(b->fd) )
		ret = 0;

noflush:
	b->fd = fd;

	return ret;
}

/**
 * @param b: a fobuf structure to use
 * @param buf: pointer to the data you want to write
 * @param len: size of data pointed to by buf
 *
 * Slow path for writing to the buffer, do not call directly instead
 * use fobuf_write().
 *
 * Return value: zero on error, non-zero on success.
 */
static int fobuf_write_slow(struct fobuf *b, const void *buf, size_t len)
{
	/* fill up the buffer before flushing, we already know
	 * that len >= b->buf_len so a full buffer flush is
	 * inevitable.
	 */
	memcpy(b->ptr, buf, b->buf_len);
	buf += b->buf_len;
	len -= b->buf_len;
	b->ptr += b->buf_len;
	b->buf_len = 0;
	if ( !__fobuf_flush(b) )
		return 0;

	/* If the remaining data is the same size as the buffer then
	 * buffering is doing an un-necessary copy of the data.
	 *
	 * If the remaining data is bigger than the buffer then we
	 * must write it out right away anyway.
	 */
	if ( len >= b->buf_sz )
		return __fd_write(b->fd, buf, len);

	/* normal write - len may be zero */
	memcpy(b->ptr, buf, len);
	b->ptr += len;
	b->buf_len -= len;

	return 1;
}
