/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * TODO:
 *  o Add date/time and log level columns
 *  o Colorize error entries
 *  o Add headers and allow them to be sorted and reorderd
 *  o Delete old entries after a while
 *  o Conserve memory using a custom tree model
*/
#include <firestorm.h>
#include <gnome.h>
#include <g_interface.h>
#include <g_mesg.h>

#define MSGLIST_ICON	0
#define MSGLIST_PRIO	1
#define MSGLIST_MESG	2
#define MSGLIST_MAX	3

static GtkListStore *msglist;
static GtkWidget *last_tree;
static const char *level[M_MAX]={
	"Unclassified",
	"Critical",
	"Error",
	"Warning",
	"Info",
	"Debug",
};

static const gchar *icon[M_MAX]={
	GTK_STOCK_DIALOG_QUESTION,
	GTK_STOCK_DIALOG_ERROR,
	GTK_STOCK_DIALOG_ERROR,
	GTK_STOCK_DIALOG_WARNING,
	GTK_STOCK_DIALOG_INFO,
	GTK_STOCK_EXECUTE,
};

static void glib_msg(const char *string)
{
	mesg(M_INFO, "%s", string);
}

static void glib_errmsg(const char *string)
{
	mesg(M_ERR, "%s", string);
}

static void glib_log(const gchar *log_domain,
			GLogLevelFlags log_level,
			const gchar *message,
			gpointer user_data)
{
	mesg_code_t code = M_INFO;

	if ( log_level & G_LOG_LEVEL_DEBUG )
		code = M_DEBUG;
	if ( log_level & G_LOG_LEVEL_INFO )
		code = M_INFO;
	if ( log_level & G_LOG_LEVEL_MESSAGE )
		code = M_INFO;
	if ( log_level & G_LOG_LEVEL_WARNING )
		code = M_WARN;
	if ( log_level & G_LOG_LEVEL_ERROR )
		code = M_ERR;
	if ( log_level & G_LOG_LEVEL_CRITICAL )
		code = M_CRIT;

	g_log_default_handler(log_domain, log_level, message, user_data);

	mesg(code, "%s: %s",
		log_domain ? log_domain : "App",
		message);
}

void g_mesg_backend(mesg_code_t code, const char *str, size_t len)
{
	GtkTreeIter i;

	if ( code > M_MAX )
		code = M_UNSET;

	if ( code == M_ERR )
		gnome_error_dialog_parented(str, GTK_WINDOW(mainwin));

	gtk_list_store_append(msglist, &i);
	gtk_list_store_set(msglist, &i,
				MSGLIST_ICON, icon[code],
				MSGLIST_PRIO, level[code],
				MSGLIST_MESG, str,
				-1);

	if ( last_tree ) {
		GtkTreePath *path;
		path = gtk_tree_model_get_path(GTK_TREE_MODEL(msglist), &i);
		gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(last_tree), path,
						NULL, FALSE, 0, 0);
		gtk_tree_path_free(path);
	}

	printf("%s: %s\n", level[code], str);
}

static void add_col(GtkWidget *tree, const char *title, int idx, int pix)
{
	GtkCellRenderer *cell;
	GtkTreeViewColumn *col;

	col = gtk_tree_view_column_new();
	gtk_tree_view_column_set_reorderable(col, TRUE);
	gtk_tree_view_column_set_sort_column_id(col, idx);

	if ( pix >= 0 ) {
		cell = gtk_cell_renderer_pixbuf_new();
		gtk_tree_view_column_pack_start(col, cell, FALSE);
		gtk_tree_view_column_set_attributes(col, cell,
						"stock-id", pix, NULL);
	}

	cell = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", idx, NULL);
	gtk_tree_view_column_set_title(col, title);
	gtk_tree_view_column_set_resizable(col, TRUE);

	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), col);
}

GtkWidget *g_mesg_view_new(void)
{
	GtkWidget *scroll;
	GtkWidget *tree;

	tree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(msglist));
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree), TRUE);
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree), FALSE);
	gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(tree), FALSE);
	add_col(tree, "Priority", MSGLIST_PRIO, MSGLIST_ICON);
	add_col(tree, "Message", MSGLIST_MESG, -1);

	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_container_add(GTK_CONTAINER(scroll), tree);

	last_tree = tree;

	return scroll;
}

int g_mesg_init(void)
{
	msglist = gtk_list_store_new(MSGLIST_MAX,
					G_TYPE_STRING,
					G_TYPE_STRING,
					G_TYPE_STRING);

	if ( msglist == NULL )
		return 0;

	g_set_print_handler(glib_msg);
	g_set_printerr_handler(glib_errmsg);

	g_log_set_handler("GLib", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL |
			G_LOG_FLAG_RECURSION, glib_log, NULL);
	g_log_set_handler("Gtk", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL |
			G_LOG_FLAG_RECURSION, glib_log, NULL);
	g_log_set_handler("GnomeUI", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL |
			G_LOG_FLAG_RECURSION, glib_log, NULL);
	g_log_set_handler(NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL |
			G_LOG_FLAG_RECURSION, glib_log, NULL);

	return 1;
}
