/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003,2004 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*
* TODO:
*  o Use shifts instead of modulo for top table
*  o Make prime-number sized secondary tables
*  o To trade space for time we could use sorted buckets and
*    get a 100% load factor.
*  o Make the primary index size variable and use the reserved
*    fields in the database header.
*  o Use initmem to copy our keys and values...
*/
#include <firestorm.h>
#include <dotscara/mpool.h>
#include <dotscara/fobuf.h>
#include <dotscara/initmem.h>
#include <idx.h>
#include <i_idx.h>
#include <ctype.h>

#define SKUNK_ALIGN	4 /* bytes */

static u_int8_t paddata[SKUNK_ALIGN];

struct skunk_rec {
	size_t key_len;
	const char *key;
	struct mpool vals;
	idx_ofs_t num_vals;
	idx_hash_t hash;
	struct skunk_rec *next;
};

struct skunk_head {
	struct skunk_rec *list;
	idx_ofs_t count;
};

struct i_skunk {
	struct i_fieldref ref;
	struct mpool rec;
	struct initmem keys;
	struct skunk_head head[SKUNK_HASHSZ];
	idx_ofs_t val_count;
	idx_ofs_t all_key_len;
};

static size_t skunk_align(size_t len)
{
	return (len + SKUNK_ALIGN) & ~(SKUNK_ALIGN-1);
}

static inline int hdr_count(struct mpool *m)
{
	int ret = 0;
	struct mpool_hdr *h;

	for(h=m->slabs; h; h=h->next)
		ret++;

	return ret;
}

static inline void reverse_mpool(struct mpool *m)
{
	int cnt = hdr_count(m);
	struct mpool_hdr *a[cnt];
	struct mpool_hdr *h;
	int i = 0;

	for(h=m->slabs; h; h=h->next)
		a[i++] = h;

	m->slabs = NULL;

	for(i=0; i < cnt; i++) {
		a[i]->next = m->slabs;
		m->slabs = a[i];
	}
}

static void skunk_resolve(struct skunk_head *h, struct skunk_rec **r2)
{
	struct skunk_rec *r;
	struct skunk_rec *list = h->list;
	unsigned int i;

	if ( h->count == 0 )
		return;

	memset(r2, 0, h->count * sizeof(*r2));

	for(r=list; r; r=r->next) {
		i = (r->hash ^ r->key_len) % h->count;
probe:
		if ( r2[i] ) {
			i = (i+1) % h->count;
			goto probe;
		}else{
			r2[i] = r;
		}
	}

	for(h->list = NULL, i = h->count - 1; ; i--) {
		if ( r2[i] ) {
			r2[i]->next = h->list;
			h->list = r2[i];
		}
		if ( i == 0 )
			break;
	}
}

static inline void size_bucket(struct skunk_head *h)
{
	if ( h->count == 0 )
		return;

	h->count += h->count / 2;

	/* TODO: make prime sized at some point */
	h->count |= 1;
}

static struct skunk_rec *find_existing(struct i_skunk *db, idx_hash_t hash,
				const char *key, size_t klen)
{
	struct skunk_head *h = &db->head[hash % SKUNK_HASHSZ];
	struct skunk_rec *r;

	for(r=h->list; r; r=r->next) {
		if ( r->hash != hash )
			continue;
		if ( r->key_len != klen )
			continue;
		if ( memcmp(r->key, key, klen) )
			continue;

		return r;
	}

	return NULL;
}

/* skunk_write:
 * @db: a valid fieldref
 * @o: an output file buffer
 *
 * Writes the database out to the file
 *
 * Return value: zero on failure, non-zero on success
 */
static int skunk_write(struct i_fieldref *f, struct fobuf *o)
{
	struct i_skunk *db = (struct i_skunk *)f;
	struct skunk_rec **table;
	struct skunk_rec *r;
	struct skunk_head *h;
	size_t max = 0;
	idx_ofs_t cur, cur_v;
	unsigned int i, j;

	cur = SKUNK_HASHSZ * sizeof(struct skunk_table);

	/* Write out the primary hash table */
	for(i=0; i < SKUNK_HASHSZ; i++) {
		struct skunk_table t;
		h = &db->head[i];

		if ( !h->count ) {
			t.ofs = 0;
			t.len = 0;
			if ( !fobuf_write(o, &t, sizeof(t)) )
				return 0;
			continue;
		}

		if ( h->count > max )
			max = h->count;

		t.ofs = cur;
		t.len = h->count;
		cur += t.len * sizeof(struct skunk_bucket);

		if ( !fobuf_write(o, &t, sizeof(t)) )
			return 0;
	}

	if ( max == 0 )
		return 1;

	table = calloc(max, sizeof(*table));
	if ( table == NULL )
		return 0;

	cur_v = cur + db->all_key_len;

	for(i=0; i < SKUNK_HASHSZ; i++) {
		struct skunk_bucket b;

		h = &db->head[i];

		skunk_resolve(h, table);

		for(j=0; j < h->count; j++) {
			if ( (r=table[j]) ) {
				b.hash = r->hash;
				b.kofs = cur;
				b.klen = r->key_len;
				b.vofs = cur_v;
				b.vlen = r->num_vals;
				cur += skunk_align(r->key_len);
				cur_v += r->num_vals * sizeof(idx_val_t);

				if ( (b.kofs & 0x3) || (b.vofs & 0x3) ) {
					mesg(M_CRIT, "eek align %u %u",
						b.kofs, b.vofs);
				}
			}else{
				memset(&b, 0, sizeof(b));
			}

			if ( !fobuf_write(o, &b, sizeof(b)) ) {
				free(table);
				return 0;
			}
		}
	}

	free(table);

	/* Write out the keys */
	for(i=0; i < SKUNK_HASHSZ; i++) {
		h = &db->head[i];

		for(r=h->list; r; r=r->next) {
			size_t klen = r->key_len;
			size_t rlen = skunk_align(klen);

			if ( !fobuf_write(o, r->key, klen) )
				return 0;
			if ( rlen == klen )
				continue;
			if ( !fobuf_write(o, paddata, rlen-klen) )
				return 0;
		}
	}

	/* Write out the values */
	for(i=0; i < SKUNK_HASHSZ; i++) {
		h = &db->head[i];

		for(r=h->list; r; r=r->next) {
			struct mpool_hdr *hdr;
			size_t tlen = 0;

			for(hdr=r->vals.slabs; hdr; hdr = hdr->next) {
				size_t len;

				len = (char *)hdr->next_obj - (char *)hdr;
				if ( len > r->vals.slab_size )
					len -= r->vals.obj_size;

				len -= sizeof(*hdr);
				tlen += len;

				if ( !fobuf_write(o, hdr->data, len) )
					return 0;
			}

			if ( tlen != (r->num_vals * sizeof(idx_val_t)) )
				return 0;
		}
	}

	return 1;
}

/* skunk_insert_str:
 * @db: a valid fieldref
 * @key: key to use
 * @klen: length of key
 * @val: offset in file being indexed
 *
 * Inserts an item in to the index.
 *
 * Return value: non-zero on success, zero on failure
 */
static int skunk_insert_str(struct i_fieldref *f, const char *key,
				size_t klen, idx_val_t val)
{
	struct i_skunk *db = (struct i_skunk *)f;
	struct skunk_rec *r;
	struct skunk_head *h;
	idx_hash_t hash;
	idx_val_t *the_val;

	hash = skunk_hash(key, klen);
	r = find_existing(db, hash, key, klen);
	if ( r )
		goto just_append_val;

	r = mpool_alloc(&db->rec);
	if ( r == NULL )
		return 0;

	r->key = key;
	r->key_len = klen;
	mpool_new(&r->vals, sizeof(idx_val_t), 0);
	r->num_vals = 0;
	r->hash = hash;
	r->next = NULL;


	h = &db->head[hash % SKUNK_HASHSZ];
	r->next = h->list;
	h->list = r;
	h->count++;

just_append_val:
	the_val = mpool_alloc(&r->vals);
	if ( the_val == NULL )
		return 0;

	*the_val = val;
	db->val_count++;
	r->num_vals++;

	return 1;
}

static int skunk_insert_str_nocase(struct i_fieldref *f, const char *key,
				size_t klen, idx_val_t val)
{
	struct i_skunk *db = (struct i_skunk *)f;
	char *str = initmem_alloc(&db->keys, klen);
	size_t i;

	if ( str == NULL )
		return 0;

	for(i=0; i < klen; i++)
		str[i] = toupper(key[i]);

	return skunk_insert_str(f, str, klen, val);
}

/* skunk_new:
 *
 * Creates a new DB for inserting in to.
 *
 * Return value: NULL on error
 */
static struct i_fieldref *skunk_new(void)
{
	struct i_skunk *db;

	db = calloc(1, sizeof(*db));
	if ( db == NULL )
		return NULL;

	if ( !mpool_new(&db->rec, sizeof(struct skunk_rec), 0) ) {
		free(db);
		return NULL;
	}

	if ( !initmem_new(&db->keys, 0) ) {
		mpool_free(&db->rec);
		free(db);
		return NULL;
	}

	return (struct i_fieldref *)db;
}

/* skunk_free:
 * @db: a valid fieldref
 *
 * Deletes all data currently associated with the database.
 */
static void skunk_free(struct i_fieldref *f)
{
	struct i_skunk *db = (struct i_skunk *)f;
	struct skunk_rec *r;
	unsigned int i;

	for(i=0; i < SKUNK_HASHSZ; i++) {
		for(r=db->head[i].list; r; r=r->next) {
			mpool_free(&r->vals);
		}
	}
	mpool_free(&db->rec);
	initmem_free(&db->keys);
	free(db);
}

/* skunk_prepare:
 * @db: a valid #i_ordered database
 * @size: pointer to size of total DB
 * @keys: pointer to number of keys
 * @vals: pointer to number of vals
 *
 * This method is called once after all inserts have been done, but
 * before writing. The size, keys and vals params must all be filled
 * in in a valid way.
 *
 * Return value: zero on success, less than zero for error.
 */
static int skunk_prepare(struct i_fieldref *f, idx_ofs_t *size,
			idx_ofs_t *keys, idx_ofs_t *vals)
{
	struct i_skunk *db = (struct i_skunk *)f;
	struct skunk_rec *r;
	idx_ofs_t k = 0;
	idx_ofs_t sz = 0;
	unsigned int i;

	sz += SKUNK_HASHSZ * sizeof(struct skunk_table);
	sz += db->val_count * sizeof(idx_val_t);

	db->all_key_len = 0;

	for(i=0; i < SKUNK_HASHSZ; i++) {
		size_bucket(&db->head[i]);

		sz += sizeof(struct skunk_bucket) * db->head[i].count;
		k += db->head[i].count;

		for(r = db->head[i].list; r; r = r->next) {
			size_t ksz;
			reverse_mpool(&r->vals);
			ksz = skunk_align(r->key_len);
			sz += ksz;
			db->all_key_len += ksz;
		}
	}

	*size = sz;
	*keys = k;
	*vals = db->val_count;

	return 1;
}


/* String type */
struct i_type idx_skunk __private = {
	.id = IDX_SKUNK,
	.new = skunk_new,
	.insert = NULL,
	.insert_str = skunk_insert_str,
	.prepare = skunk_prepare,
	.write = skunk_write,
	.free = skunk_free,
};

/* Case-insensitive string type */
struct i_type idx_skunk_nocase __private = {
	.id = IDX_SKUNK_NOCASE,
	.new = skunk_new,
	.insert = NULL,
	.insert_str = skunk_insert_str_nocase,
	.prepare = skunk_prepare,
	.write = skunk_write,
	.free = skunk_free,
};
