/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2004 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
 *
 * Functions:
 *  o eventloop_find() - Find an eventloop plugin by name
 *  o eventloop_add() - Register an eventloop plugin
 *  o eventloop_load() - Load all nbio plugins
 *  o nbio_init() - Initialise nbio
 *  o nbio_fini() - Deinitialise nbio
 *  o nbio_pump() - Pump events
 *  o nbio_add() - Register an fd with read/write/error callbacks
 *  o nbio_del() - Remove an fd
*/
#include <firestorm.h>
#include <loader.h>
#include <nbio.h>

static struct eventloop *ev_list;

struct eventloop *eventloop_find(const char *name)
{
	struct eventloop *e;

	for(e=ev_list; e; e=e->next)
		if ( !strcmp(e->name, name) )
			break;

	return e;
}

static void eventloop_add(struct plugin *p, struct eventloop *e)
{
	for(; e->name; e++) {
		if ( eventloop_find(e->name) ) {
			mesg(M_ERR, "eventloop: '%s' eventloop is "
				"already registered", e->name);
			continue;
		}
		e->next = ev_list;
		ev_list = e;
	}
}

void eventloop_load(void)
{
	loader_find_syms("firestorm_eventloops", (loader_cb_t)eventloop_add);
}

int nbio_init(struct iothread *t, const char *plugin)
{
	if ( plugin == NULL ) {
		t->plugin = ev_list;
		if ( t->plugin == NULL )
			firestorm_exit(EXIT_FAILURE, "No eventloop plugins");
	}else{
		t->plugin = eventloop_find(plugin);
	}

	if ( t->plugin == NULL )
		return 0;

try_again:
	if ( !t->plugin->init(t) ) {
		if ( plugin == NULL && t->plugin->next ) {
			t->plugin = t->plugin->next;
			goto try_again;
		}
		return 0;
	}

	INIT_LIST_HEAD(&t->active);
	INIT_LIST_HEAD(&t->inactive);
	return 1;
}

void nbio_fini(struct iothread *t)
{
	struct nbio *n, *tmp;

	list_for_each_entry_safe(n, tmp, &t->inactive, list) {
		list_move_tail(&n->list, &t->active);
		t->plugin->active(t, n);
	}

	list_for_each_entry_safe(n, tmp, &t->active, list)
		n->ops->dtor(n);

	t->plugin->fini(t);
}

void nbio_pump(struct iothread *t)
{
	struct nbio *n, *tmp;

	t->plugin->pump(t);

	list_for_each_entry_safe(n, tmp, &t->active, list) {
		unsigned short ev = n->flags & n->mask;

		if ( ev & NBIO_READ )
			n->ops->read(n);
		if ( ev & NBIO_WRITE )
			n->ops->write(n);
	}
}

void nbio_add(struct iothread *t, struct nbio *n)
{
	list_add(&n->list, &t->active);
	t->plugin->active(t, n);
}

void nbio_del(struct iothread *t, struct nbio *n)
{
	t->plugin->active(t, n);
	list_del(&n->list);
	n->ops->dtor(n);
}

void nbio_inactive(struct iothread *t, struct nbio *n)
{
	list_move(&n->list, &t->inactive);
	t->plugin->inactive(t, n);
}
