/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * mesg backend for the sensor, we just print to stdout.
 */
#include <firestorm.h>
#include <dotscara/fdctl.h>

#include <fcntl.h>
#include <unistd.h>

#include <sensor.h>
#include <elog.h>

static char *code_str[] = {
	"undefined","critical","error",
	"warning","info","debug",
};

#define MESG_INITBUF_LEN 8192
static int buffering = 1;
static size_t abuflen, ibuflen;
static char *abuf, *ibuf;

/* print out messages stored in the internal buffer */
static void smesg_flush(struct packet *pkt, int l, void *priv)
{
	if ( !buffering )
		return;

	printf("%s", ibuf);

	free(ibuf);
	buffering = 0;
}

/* Initialise the logging code */
void smesg_init(void)
{
	notify_exit(smesg_flush, NULL);
}

/* Print a message to the internal text buffer */
/* We know at this point how the user wants us to log
 * from the config file, so lets do it */
void smesg_logfile(const char *fn)
{
	int fd;

	fflush(stdout);

	if ( !fn )
		goto no_open;

	if ( (fd=open(fn, O_WRONLY|O_CREAT|O_TRUNC, 00640))<0 ) {
		firestorm_exit(EXIT_FAILURE, "%s: open(): %s", fn, sys_err());
	}

	if ( fd!=1 ) {
		if ( dup2(fd, 1)<0 )
			firestorm_exit(EXIT_FAILURE, "%s: dup2(): %s", fn, sys_err());
		fd_close(fd);
	}

no_open:
	smesg_flush(NULL, -1,NULL);
}

/* put a message in to the initial buffer */
static void smesg_initlog(struct timeval *tv, int code, const char *buf)
{
	int len;
	char *new;
	size_t new_len;

again:
	len = snprintf(abuf, abuflen,
			"%.8lu.%.6lu %s: %s\n",
			tv->tv_sec, tv->tv_usec, code_str[code], buf);
	if ( len < 0 ) /* bug in old glibc */
		len = 1;
	if ( (size_t)len < abuflen ) {
		abuflen -= len;
		abuf += len;
		return;
	}

	new_len =  ibuflen + (len - abuflen) + 1;
	new = realloc(ibuf, new_len);
	if ( new == NULL ) {
		*abuf = '\0';
		return;
	}

	ibuf = new;
	abuf = new + (ibuflen - abuflen);
	abuflen = abuflen + (new_len - ibuflen);
	ibuflen = new_len;
	goto again;
}

/* Print out a firestorm internal system log message */
void smesg_backend(mesg_code_t code, const char *str, size_t len)
{
	struct timeval tv;

	if ( code > M_MAX )
		code = 0;

	gettimeofday(&tv, NULL);

	if ( !buffering ) {
		/* No point checking for errors, what are we gonna do
		 * print them to screen? ;) */
		printf("%.8lu.%.6lu %s: %s\n", tv.tv_sec,
			tv.tv_usec, code_str[code], str);
		fflush(stdout);
		return;
	}else{
		smesg_initlog(&tv, code, str);
	}
}
