/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * Capdev plugin which reads elog files
 *
 * KNOWN BUGS:
 *  o Whole file is mmaped, if there isn't enough contiguous virtual
 *    space, this will fail.
 *  o If the file is modified *during* use, this code may be exploited.
 *    However this is nothing new. If user A can write to file that user
 *    B is reading, user A can become user B, in all but the most trivial
 *    cases (cp/cat/etc..)
 *
 * TODO:
 *  o Make code more robust in sequential access mode
 *  o Support large files
 *  o Move to elog 0.3
 *  o add 0.2 backwards compat code
*/
#include <firestorm.h>
#include <dotscara/fdctl.h>

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <unistd.h>

#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capdev.h>
#include <nbio.h>
#include <args.h>
#include <elog.h>

struct elog_priv {
	u_int8_t *map;
	size_t maplen, ofs;
	struct packet pkt;
	struct generator gen;
	struct alert alert;
	char *fn;
	int fd;
};

static struct arg elog_args[]={
	{"file", ARGTYPE_STRING, .help = "Input filename"},
	{NULL, }
};

static void free_sessions(struct packet *pkt)
{
	int i;
	struct layer *l;

	for(i=0; i < pkt->llen; i++) {
		l = &pkt->layer[i];
		if ( l->session && l->proto && l->proto->free ) {
			l->proto->free(l->session);
		}
	}

	pkt->llen = 0;
}

/* Parse alert decode data */
static int do_decode(struct packet *pkt, char *buf, int buflen)
{
	struct elog_dhdr *d;
	char *end = buf + buflen;
	struct proto *p;

	while( buf + sizeof(*d) <= end) {
		char *pnam, *pser, *dend;

		d=(struct elog_dhdr *)buf;

		/* Check the validity */
		if ( ((size_t)d->tot_len<<2) < sizeof(*d) )
			return 0;
		if ( ((size_t)d->tot_len<<2) < (d->name_len<<2) + sizeof(*d) )
			return 0;
		if ( buf + (d->tot_len<<2) > end )
			return 0;

		/* Work out where stuff is */
		pnam=(buf + sizeof(*d));
		pser = pnam+(d->name_len<<2);
		dend = buf+(d->tot_len<<2);

		if ( d->name_len == 0 ) {
			p = NULL;
		}else{
			p = decode_proto(pnam);
		}

		pkt->layer[pkt->llen].h.raw =
			pkt->base + sys_be16(d->pkt_ofs);
		pkt->layer[pkt->llen].proto = p;
		pkt->layer[pkt->llen].flags = sys_be32(d->flags);

		if ( p && p->deserialize && dend > pser ) {
			pkt->layer[pkt->llen].session = p->deserialize(
				pkt, pkt->llen, pser, dend - pser);
		}else{
			pkt->layer[pkt->llen].session = NULL;
		}

		if ( pkt->llen++ >= PKT_LAYERS )
			return 1;

		buf += d->tot_len << 2;
	}

	return 1;
}

static struct packet *elog_query(struct source *s, off_t ofs)
{
	struct elog_priv *p = s->priv;
	struct elog_pkthdr *ph;
	struct timeval tv;
	u_int8_t *decode_buf, *pkt_data, *str_g, *str_a;
	u_int8_t *cur, *end;

	if ( (p->ofs + sizeof(*ph)) > p->maplen )
		return NULL;

	cur = p->map + ofs;
	end = p->map + p->maplen;
	ph = (struct elog_pkthdr *)cur;
	decode_buf = cur + sizeof(*ph);
	pkt_data = decode_buf + (sys_be16(ph->decode_len) << 2);
	str_g = pkt_data + (sys_be32(ph->pkt_caplen));
	str_a = str_g + ph->gen_len;

	/* Integrity checks */
	if ( ph->h.type != sys_be16(ELOG_ALERT) )
		return NULL;
	if ( ph->h.reclen < sys_be32(sizeof(*ph)) )
		return NULL;
	if ( ofs + sys_be32(ph->h.reclen) > p->maplen )
		return NULL;
	if ( str_a + ph->alert_len > end )
		return NULL;

	/* XXX: In seqeuntial mode, such packets may be safely skipped */
	if ( str_g[ph->gen_len-1] ||
		str_a[ph->alert_len-1] )
		return NULL;

	p->gen.name = str_g;
	p->gen.t = NULL;
	p->alert.alert = str_a;
	p->alert.sid = sys_be32(ph->sid);
	p->alert.rev = sys_be32(ph->rev);
	p->alert.priority = ph->h.prio;
	p->alert.gen = &p->gen;

	p->pkt.len = sys_be32(ph->pkt_len);
	p->pkt.caplen = sys_be32(ph->pkt_caplen);
	p->pkt.llen = 0;
	p->pkt.base = pkt_data;
	p->pkt.end = pkt_data + p->pkt.caplen;
	tv.tv_sec = sys_be32(ph->h.tv_sec);
	tv.tv_usec = sys_be32(ph->h.tv_usec);
	p->pkt.time = time_from_timeval(&tv);
	p->pkt.flags = sys_be32(ph->pflags) & FPMASK_PERSISTANT;
	p->pkt.alert = &p->alert;

	/* XXX: May also be skipped in sequential mode */
	if ( !do_decode(&p->pkt, decode_buf, sys_be16(ph->decode_len)<<2) )
		return NULL;

	return &p->pkt;
}

static void elog_rewind(struct source *s)
{
	struct elog_priv *p = s->priv;
	p->ofs = sizeof(struct efile_hdr);
}

static int elog_init(struct source *s, const char *args)
{
	struct elog_priv *p;
	struct stat st;
	struct fvec *fnvec;
	struct efile_hdr *fh;

	switch ( args_parse(elog_args, args, NULL) ) {
	case -1:
		mesg(M_ERR, "elog: parse error: %s", args);
	case 0:
		return 0;
	default:
		break;
	}

	p = calloc(1, sizeof(*p));
	if ( p == NULL )
		return 0;

	/* Copy filename */
	fnvec = &elog_args[0].val.v_str;
	p->fn = malloc(fnvec->len + 1);
	if ( p->fn == NULL )
		goto err_free;
	memcpy(p->fn, fnvec->ptr, fnvec->len);
	p->fn[fnvec->len] = '\0';

	p->fd = open(p->fn, O_RDONLY);
	if ( p->fd < 0 ) {
		mesg(M_ERR, "elog: %s: open(): %s", p->fn, sys_err());
		goto err_free_fn;
	}

	if ( fstat(p->fd, &st) ) {
		mesg(M_ERR, "elog: %s: open(): %s", p->fn, sys_err());
		goto err_close;
	}

	if ( st.st_size > SIZE_MAX ) {
		mesg(M_ERR, "elog: %s: Sorry, file is too large", p->fn);
		goto err_close;
	}

	p->maplen = (size_t)st.st_size;
	p->map = mmap(NULL, p->maplen, PROT_READ, MAP_SHARED, p->fd, 0);
	if ( p->map == MAP_FAILED ) {
		mesg(M_ERR, "elog: %s: mmap(): %s", p->fn, sys_err());
		goto err_close;
	}

	fh = (struct efile_hdr *)p->map;
	if ( p->maplen < sizeof(*fh) ||
		fh->magic != sys_be32(EF_MAGIC) ||
		fh->vers_major != EF_VERS_MAJ ||
		fh->vers_minor != EF_VERS_MIN ) {
		mesg(M_ERR, "elog: %s: Not a valid elog file", p->fn);
		goto err_unmap;
	}

#if HAVE_MADVISE && defined(MADV_SEQUENTIAL)
	madvise(p->map, p->maplen, MADV_SEQUENTIAL);
#endif

	capdev_init_packet(&p->pkt, s);

	s->priv = p;
	elog_rewind(s);

	return 1;

err_unmap:
	munmap(p->map, p->maplen);
err_close:
	fd_close(p->fd);
err_free_fn:
	free(p->fn);
err_free:
	free(p);
	return 0;
}

/* Stop the capture and cleanup our process */
static void elog_free(struct source *s)
{
	struct elog_priv *p = s->priv;

	if ( p->pkt.llen )
		free_sessions(s->last_pkt);

	munmap(p->map, p->maplen);
	fd_close(p->fd);
	free(p->fn);
	free(p);
}

static struct packet *elog_dequeue(struct source *s)
{
	struct elog_priv *p = s->priv;
	struct elog_common_hdr *h;
	struct packet *ret;

	if ( p->pkt.llen )
		free_sessions(s->last_pkt);

again:
	if ( (p->ofs + sizeof(*h)) > p->maplen )
		return NULL;

	h = (struct elog_common_hdr *)(p->map + p->ofs);
	if ( h->type != sys_be16(ELOG_ALERT) ) {
		p->ofs += sys_be32(h->reclen);
		goto again;
	}

	ret = elog_query(s, p->ofs);
	if ( ret )
		p->ofs += sys_be32(h->reclen);

	return ret;
}

struct capdev firestorm_capdevs[]={
	{
		.name = "elog",
		.desc = "Firestorm Enhanced Log (elog) Reader",
		.init = elog_init,
		.free = elog_free,
		.dequeue = elog_dequeue,
		.rewind = elog_rewind,
		.query = elog_query,
		.args = elog_args,
	},
	{ .name = NULL }
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "capdev.elog",
	.desc		= "Reads packets from elog files",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 0,
};
